# -*- coding: utf-8 -*-

import warnings
from tkinter import END

import matplotlib.pyplot as plt
import numpy as np

import ..putils.averages as averages
from .baseplot import Plotting
from .diagnostic import Diagnostic
from ..putils.derivatives import derivativeO1
from scipy import integrate

class DiagBallamp(Diagnostic):
    def __init__(self, avail_vars=None, specnames=None):
        super().__init__()
        self.name = 'Ballooning representation'
        self.tabs = ['fluxtube']
        
        self.help_txt = """Plot the ballooning structure of fields
                        \n
                        \nPlot real/imag : plot real and imaginary part (def. True)
                        \nNormalize :  normalize with max amplitude (def. False)
                        \nky ind : index of the mode to plot (def. first non zero)
                        \nt avg : time averaged values (def. False)
                        """

        self.avail_vars = avail_vars
        self.specnames = specnames

        self.opts = {'re_im': {'tag': "Plot real/imag", 'values': [True, False]},
                     'norm': {'tag': "Normalize", 'values': [False, True]},
                     't_avg': {'tag': "t avg", 'values': [False, True]},
                     'kyind': {'tag': "ky ind", 'values': None}}
        self.set_defaults()                     

    def get_kperp(self,run):
        nx = run.parameters.pnt.nx0
        ikx_grid = np.arange(- nx // 2 + 1, nx // 2 + 1)
        ky = float(run.parameters.pnt.kymin)
        dkx = 2. * np.pi * float(run.parameters.pnt.shat) * float(ky)

        nz = int(run.parameters.pnt.nz0)
        lx = float(run.parameters.pnt.lx)
        dkx = 2. * np.pi * float(run.parameters.pnt.shat) * float(ky)

        #print('\n\n\n\n\n\n\n')
        #print('kx_center' in run.parameters.pnt)
        #print(type(run.parameters.pnt))
        #print('\n\n\n\n\n\n\n')
        if run.parameters.pnt.kx_center:
            kx_center = run.parameters.pnt.kx_center
        else:
            kx_center = 0.

        self.kperp = np.zeros(nx*nz,dtype='float128')

        #print('kx_center',kx_center)
        #print('run.parameters.pnt.kx_center',run.parameters.pnt.kx_center)
        for i in ikx_grid:
            kx = i*dkx+kx_center
            #print('kx',kx)
            this_kperp = np.sqrt(self.geom.gxx*kx**2+2.*self.geom.gxy*kx*ky+self.geom.gyy*ky**2)
            self.kperp[(i-ikx_grid[0])*nz:(i-ikx_grid[0]+1)*nz]=this_kperp

    def get_omd(self,run,which_field = 0,mass_ratio = 1,charge = 1,plot = False):
        #Note: mass ratio only affects the kperp normalization for the Bessel functions
        #Default field is phi
        nx = run.parameters.pnt.nx0
        ikx_grid = np.arange(- nx // 2 + 1, nx // 2 + 1)
        nz = run.parameters.pnt.nz0
        lx = run.parameters.pnt.lx
        ky = run.parameters.pnt.kymin
        dkx = 2. * np.pi * run.parameters.pnt.shat * ky
    
        total_omnt = 0
        for i in run.parameters.species:
            total_omnt += i['omn']+i['omt']     
        dpdx_tot = run.parameters.pnt.beta * total_omnt
                ##print 'Debug: ', 'dpdx_pm =', pars['dpdx_pm'], \
        ##      'should be equal to beta*(omni+omti+omne+omte) =', dpdx_tot
    
        if run.parameters.pnt.kx_center:
            kx_center = run.parameters.pnt.kx_center
        else:
            kx_center = 0.

        omd_curv = np.zeros(nx*nz,dtype='float128')
        omd_gradB = np.zeros(nx*nz,dtype='float128')
    
        for i in ikx_grid:
            kx = i*dkx+kx_center
            this_omegad_gradB = -(self.geom.Kx*kx+self.geom.Ky*ky)/self.geom.Bfield
            this_omegad_curv = this_omegad_gradB + \
                               ky * float(run.parameters.pnt.dpdx_pm)/self.geom.Bfield**2/2.
            #this_omegad_curv = 2.*this_omegad
    
            omd_curv[(i-ikx_grid[0])*nz:(i-ikx_grid[0]+1)*nz]=this_omegad_curv
            omd_gradB[(i-ikx_grid[0])*nz:(i-ikx_grid[0]+1)*nz]=this_omegad_gradB
            #plt.plot(self.geom.Kx)
            #plt.plot(self.geom.Ky)
            #plt.show()
    
        self.omd_curv = omd_curv
        self.omd_gradB = omd_gradB

        #print(run.parameters.pnt.magn_geometry)
        if run.parameters.pnt.magn_geometry == 's_alpha' or run.parameters.pnt.magn_geometry == 'slab':
            a=input("WARNING!!! omd calculation has not been verified for s_alpha or slab!!!!! (press any key to continue)")
        #    if run.parameters.pnt.amhd:
        #        amhd = pars['amhd']
        #    else:
        #        amhd = 0.
        #    z_grid = np.linspace(-np.pi,np.pi, nz, endpoint = False)
        #    Kx0 = -np.sin(z_grid)/run.parameters.pnt.major_R
        #    Ky0 = -(np.cos(z_grid)+np.sin(z_grid)*\
        #          (run.parameters.pnt.shat*z_grid-amhd*\
        #          np.sin(z_grid)))/run.parameters.pnt.major_R
        #    omega_d0 = -(Kx0*kx_center+Ky0*ky)
        #    omega_d00 = omega_d0+amhd/run.parameters.pnt.q0**2/run.parameters.pnt.major_R/2.*ky/run.parameters.pnt.gBfield**2
        #    gxx0 = 1.
        #    gxy0 = run.parameters.pnt.shat*z_grid-amhd*np.sin(z_grid)
        #    gyy0 = 1+(run.parameters.pnt.shat*z_grid-amhd*np.sin(z_grid))**2
        #    kperp0 = np.sqrt(gxx0*kx_center**2+2.*gxy0*kx_center*ky+gyy0*ky**2)
        #plot=True
        if plot:
            plt.plot(self.kperp,label='kperp')
            plt.title('entire simulation domain')
            plt.legend()
            plt.show()
            plt.plot(omd_curv,label='omd_curv')
            plt.plot(omd_gradB,label='omd_gradB')
            plt.title('entire simulation domain')
            plt.legend()
            plt.show()
            #np.savetxt('omd_curv_gradB',np.column_stack((self.chi_ext,omd_curv,omd_gradB)))

        avg_omd_curv = self.eigenfunction_average_bessel(self.omd_curv,self.ballamps[0][which_field,:],run,mass_ratio=mass_ratio,charge = charge)
        avg_omd_gradB = self.eigenfunction_average_bessel(self.omd_gradB,self.ballamps[0][which_field,:],run,mass_ratio=mass_ratio,charge = charge)
        return avg_omd_curv,avg_omd_gradB

    def dfielddz(self,run,which_field=0):
        field = self.ballamps[0][which_field,:]
        dz = self.chi_ext[1]-self.chi_ext[0]  
        dfielddz = derivativeO1(dz,field,4)
        dfielddz = dfielddz / self.jacxB_ext
        #plt.plot(self.chi_ext,dfielddz)
        #plt.show()
        return dfielddz

    def eigenfunction_average_bessel(self,quantity,field,run, mass_ratio = 1,charge = 1, field_weighted = True, zstart = 0, zend = -1):
        kperp = self.kperp*np.sqrt(mass_ratio)/abs(charge)
        alpha = 2./3.
        bessel_factor = 1. / np.sqrt(1. + 2. * (kperp**2 + np.pi * alpha * kperp**4) /
                    (1. + alpha * kperp**2))
        show_plots = False
        if mass_ratio == 1 and show_plots:
            plt.plot(bessel_factor)
            plt.title('bessel_factor')
            plt.show()
            plt.plot(self.jacxB_ext)
            plt.title('jacxB_ext')
            plt.show()
            np.savetxt('bessel_jacxB.dat',np.column_stack((self.chi_ext,bessel_factor,self.jacxB_ext)))
        sum_ddz = 0
        denom = 0
        if zend == -1:
            zend = len(self.chi_ext) - 1
        if field_weighted:
            weight = field
        else:
            weight = np.ones_like(field)

        sum_ddz = 0.5*integrate.simps(quantity[zstart:zend] *abs(weight[zstart:zend])**2* bessel_factor[zstart:zend]*self.jacxB_ext[zstart:zend],self.chi_ext[zstart:zend])
        denom =  0.5*integrate.simps(abs(field[zstart:zend])**2 * bessel_factor[zstart:zend] *self.jacxB_ext[zstart:zend],self.chi_ext[zstart:zend])
        #print('sum_ddz',sum_ddz)
        #print('denom',denom)
        avg_quantity = sum_ddz/denom
        return avg_quantity

    def eigenfunction_average_bessel_old(self,quantity,field,run, mass_ratio = 1,charge = 1,field_weighted = True, zstart = 0, zend = -1):
        kperp = self.kperp*np.sqrt(mass_ratio)/abs(charge)
        alpha = 2./3.
        bessel_factor = 1. / np.sqrt(1. + 2. * (kperp**2 + np.pi * alpha * kperp**4) /
                    (1. + alpha * kperp**2))
        sum_ddz = 0
        denom = 0
        if zend == -1:
            zend = len(self.chi_ext) - 2
        if field_weighted:
            weight = field
        else:
            weight = np.ones_like(field)
        for i in range(zstart,zend):
            sum_ddz = sum_ddz + 0.5*((quantity[i]) *abs(weight[i])**2* bessel_factor[i] \
                  + quantity[i+1] * abs(weight[i+1])**2*bessel_factor[i+1])*\
                  (self.chi_ext[i+1]-self.chi_ext[i])*self.jacxB_ext[i]
            denom = denom + 0.5*(abs(field[i])**2 * bessel_factor[i] \
                  + abs(field[i+1])**2* bessel_factor[i+1])*\
                  (self.chi_ext[i+1]-self.chi_ext[i])*self.jacxB_ext[i]
        #print('sum_ddz',sum_ddz)
        #print('denom',denom)
        avg_quantity = sum_ddz/denom
        return avg_quantity

    def get_kpar(self,run,which_field = 0,mass_ratio = 1,charge = 1):
        #note: mass ratio is species mass over reference

        field = self.ballamps[0][which_field,:]
        imax = np.argmax(abs(field))
        field0 = field / abs(field[imax])

        dfielddz = self.dfielddz(run,which_field = which_field)
        dfielddz = dfielddz / abs(field[imax])
        avg_kz = np.sqrt(self.eigenfunction_average_bessel(abs(dfielddz)**2,field0,run,mass_ratio=mass_ratio,charge=charge,field_weighted = False))

        avg_kz = avg_kz * np.sqrt(1/mass_ratio)
        return avg_kz

    def calc_epar(self,run,show_plots = False):
        gradphi = self.dfielddz(run)
        apar = self.ballamps[0][1,:]
        #print(self.opts)
        if self.opts['norm']['value']:
            raise RuntimeError("Can't calculate E parallel with the normalized option.")
        if run.omega and run.gamma and run.omega != 0 and run.gamma != 0:
            omega_complex = run.omega*1J + run.gamma
        else:
            raise RuntimeError("Can't calculate E parallel without gamma and omega")
        #print('omega_complex',omega_complex)
        if show_plots:
            plt.plot(self.chi_ext,np.real(gradphi),'--',color = 'red')
            plt.plot(self.chi_ext,np.imag(gradphi),color = 'red')
            plt.plot(self.chi_ext,np.real(-omega_complex*apar),'--',color = 'black')
            plt.plot(self.chi_ext,np.imag(-omega_complex*apar),color = 'black')
            plt.show()
        diff = np.sum(np.abs(gradphi+omega_complex*apar)/self.jacxB_ext)
        diff_norm = diff/(np.sum(np.abs(gradphi)/self.jacxB_ext)+np.sum(np.abs(omega_complex*apar)/self.jacxB_ext))
        print("E_par/(abs(gradphi) + abs(omega Apar))",diff_norm)
        self.epar = diff_norm

    def get_info(self):
        self.need_file={'field': True,
                        'mom': False}

        return self.need_file
    
    def setup_options(self, run):
        self.geom=run.geometry
        if self.opts['kyind']['value'] is None:
            # default: 1st finite ky
            self.kyind = (1 if run.parameters.pnt.nonlinear else 0)  
        else:
            self.kyind = self.opts['kyind']['value']
            
        self.kyval = run.spatialgrid.ky[self.kyind]    
        self.n_fields = run.parameters.pnt.n_fields
        
        # we can set the phase factor here once and for all at intialization
        Cyq0_x0 = 1
        sign_shear = (-1 if run.parameters.pnt.shat < 0 else 1)
        nexc_sign = sign_shear*run.parameters.pnt.sign_Ip_CW*run.parameters.pnt.sign_Bt_CW

        if run.parameters.pnt.adapt_lx:
            nexc = 1
        else:
            nexc = int(np.round(run.parameters.pnt.lx*run.parameters.pnt.n_pol*np.absolute(
                run.parameters.pnt.shat)*run.parameters.pnt.kymin*np.absolute(Cyq0_x0)))
        nexc *= nexc_sign
        self.nexcj = nexc*(run.parameters.pnt.ky0_ind + self.kyind)
        self.nconn = int(int(int(run.parameters.pnt.nx0 - 1)/2)/np.abs(self.nexcj))*2 + 1
        try:
            self.phasefac = (-1) ** self.nexcj*np.exp(
                2.0*np.pi*1j*run.parameters.pnt.n0_global*run.parameters.pnt.q0*(
                            run.parameters.pnt.ky0_ind + self.kyind))
        except AttributeError:
            self.phasefac = 1.0

        self.chi_ext = []
        self.jacxB_ext = []
        for ncind in range(-int(self.nconn/2), int(self.nconn/2) + 1):
            self.chi_ext = np.concatenate(
                    [self.chi_ext, (run.spatialgrid.z + ncind*2.0*np.pi)])
            self.jacxB_ext = np.concatenate([self.jacxB_ext,self.geom.jacobian*self.geom.Bfield])
        self.nz0_2=int(run.parameters.pnt.nz0/2)
        self.get_kperp(run)
        self.ballamps = []

        
    def execute(self, data, run, steps, extensions):

        def __cast_ball(data_in):
            if self.opts['norm']['value']:
                normval = np.array(data_in[0, self.kyind, self.nz0_2],dtype=np.complex128)
                normphase = np.angle(normval)
                normval = np.absolute(normval)
            else:
                normphase = 0.0
                normval = 1.0
            
            balloon = np.array([], dtype=np.complex128)
            
            for ncind in range(-int(self.nconn/2), int(self.nconn/2) + 1):
                balloon = np.concatenate([balloon, np.conj(self.phasefac) ** ncind*data_in[
                    ncind*self.nexcj, self.kyind]])
             
            return balloon*np.exp(-1j*normphase)/normval

        eigenfunction=np.zeros((self.n_fields, self.chi_ext.size), dtype=np.complex128)
        #0=phi, 1=Apar, 2=Bpar
        eigenfunction[0,:]=__cast_ball(data.field.phi(step=steps['field'], extension=extensions['field']))
        if self.n_fields>1:
            eigenfunction[1,:]=__cast_ball(data.field.A_par(step=steps['field'], extension=extensions['field']))
        if self.n_fields>2:
            eigenfunction[2,:]=__cast_ball(data.field.B_par(step=steps['field'], extension=extensions['field']))   

        self.ballamps.append(eigenfunction)

        print("#############")
        #jacxB = geom.jacobian*geom.Bfield
         
        #omega_complex = omega*1.0J + gamma
        #gradphi = 
        self.calc_epar(run)
        self.kpar = {}
        self.omd = {}
        self.kperp_avg = {}

        print(run.parameters.pnt.dpdx_term)
        if(run.parameters.pnt.dpdx_term[1:-1] != 'full_drift' and run.parameters.pnt.dpdx_term[1:-1] != 'gradB_eq_curv' and run.parameters.pnt.dpdx > 0):
           raise RuntimeError("Error! This diagnostic only works for full_drift and gradB_eq_curv at the moment (easy to fix).")
        for i in run.parameters.species:
            print(i['name'])
            print(i)
            for j in range(run.parameters.pnt.n_fields):
                if j == 0:
                    field0 = 'phi'
                elif j==1:
                    field0 = 'apar'
                elif j==2:
                    field0 = 'bpar'
                self.kpar['kpar_'+field0+'_'+i['name']] = self.get_kpar(run,which_field = j,mass_ratio = i['mass'],charge = i['charge'])
                avg_omd_curv,avg_omd_gradB = self.get_omd(run,which_field = j, mass_ratio = i['mass'],charge = i['charge'])
                print(i['name'],field0,avg_omd_curv,avg_omd_gradB)
                self.omd['omd_'+field0+'_'+i['name']] = avg_omd_curv
                self.kperp_avg['kperp_avg_'+field0+'_'+i['name']] = np.sqrt(self.eigenfunction_average_bessel((self.kperp*i['mass']**0.5/i['charge'])**2,self.ballamps[0][j,:],run,mass_ratio = i['mass'],charge = i['charge']))
                #if(run.parameters.pnt.dpdx_term == 'full_drift'):
                #    self.omd['omd_'+field0+'_'+i['name']] += avg_omd_curv
                savetext = True
                if savetext:
                    np.savetxt('omd_curv_gradB_'+i['name']+field0,np.column_stack((self.chi_ext,self.kperp,self.omd_curv,self.omd_gradB)))
                #self.omd['omd_gradB_'+field0+'_'+i['name']] = avg_omd_gradB
        print("#############")


    def plot(self, time_requested, output=None, out_folder=None):
        self.plotbase = Plotting()
        lbl=['phi','apar','bpar']
        def plot_one_quant(fig, ncols, i_c, x, y, ttl, quant):

            # lin-lin plots
            ax_lin = fig.add_subplot(2, ncols, 1 + i_c)
            ax_lin.plot(x, np.absolute(y), label="|{}|".format(self.plotbase.titles[quant]),
                        color='black')
            if self.opts['re_im']['value']:
                ax_lin.plot(x, np.real(y), label="Re({})".format(self.plotbase.titles[quant]),
                            color='blue', lw=1)
                ax_lin.plot(x, np.imag(y), label="Im({})".format(self.plotbase.titles[quant]),
                            color='red', lw=1)
            ax_lin.set_xlabel(r"$\chi/\pi$")
            ax_lin.set_title(ttl)
            ax_lin.axhline(y=0, color='black', lw=1, ls='--')
            ax_lin.legend()

            # log-log plots
            ax_log = fig.add_subplot(2, ncols, ncols+1+i_c)
            ax_log.plot(x, np.absolute(y), label="|{}|".format(self.plotbase.titles[quant]),
                        color='black')
            if self.opts['re_im']['value']:
                ax_log.plot(x, np.real(y), label="Re({})".format(self.plotbase.titles[quant]),
                            color='blue', lw=1)
                ax_log.plot(x, np.imag(y), label="Im({})".format(self.plotbase.titles[quant]),
                            color='red', lw=1)
            ax_log.set_xlabel(r"$\chi/\pi$")
            ax_log.axhline(y=0, color='black', lw=1, ls='--')
            ax_log.set_yscale('log')
            ax_log.legend()

        chi_pi = self.chi_ext/np.pi
        
        ttl_ky = r"$k_y\rho={:6.3f} $".format(self.kyval)

        if self.opts['t_avg']['value']:
            fig = plt.figure()
            for i_q, quant in self.n_fields:
                plot_one_quant(fig, self.n_fields, i_q, chi_pi,
                             averages.mytrapz(np.asarray(self.ballamps)[:,i_q,:], time_requested), ttl_ky,
                             self.plotbase.titles[i_q])
            fig.show()


        elif len(time_requested) < 12:
            # allow max 12 spearate figures
            for i_t, time in enumerate(time_requested):
                fig = plt.figure()
                for i_q in np.arange(self.n_fields):
                    plot_one_quant(fig, self.n_fields, i_q, chi_pi, self.ballamps[i_t][i_q,:],
                                     ttl_ky + "@ t={:.3f}".format(time),lbl[i_q])
                fig.show()

        else:
            warnings.warn("Too many timesteps selected", RuntimeWarning)
            if output:
                output.info_txt.insert(END, "Too many timesteps selected\n")
                output.info_txt.see(END)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
@author: Austin Blackmon, Dongyang Kuang
"""


import sys

import numpy as np
import optparse as op
import matplotlib.pyplot as plt

from sys import path
from sys import exit
import os
import glob
import base64

from .fieldlib import fieldfile
from .ParIO import Parameters
# from .finite_differences import *
# import mgkdb.support.pydiag.utils.comm as comm
from .pydiag.utils import comm 
from .pydiag.data import datafiles

from .pydiag.utils.gkdb import GKDB_linear, GKDB_nonlin
from .putils.loader import Loader
from .data.data import Data
#from .putils.geom import Geometry
from .putils.run import Run
from .putils.simulation import Simulation
#from .putils.spatial_grid import SpatialGrid
#from .putils.vsp_grid import VspGrid
#from .data.base_file import createGENEfile

from .diagnostics.diag_flux_spectra import DiagFluxSpectra
from .diagnostics.diag_amplitude_spectra import DiagAmplitudeSpectra
from .diagnostics.diag_field_mom_snapshots import DiagFieldMomSnapshots
#=======================================================

def get_nspec(out_dir,suffix):
    #grab parameters dictionary from ParIO.py - Parameters()
    par = Parameters()
    par.Read_Pars(os.path.join(out_dir, 'parameters' + suffix))
    pars = par.pardict 
    
    #find 'n_spec' value in parameters dictionary
    nspec = pars['n_spec']
    
    return(nspec)
    
def get_nrg(out_dir, suffix):
    #modified from IFSedge/get_nrg.py
    par = Parameters()
    par.Read_Pars(os.path.join(out_dir, 'parameters' + suffix))
    pars = par.pardict 
    
    #initializations
    ncols=pars['nrgcols']
    time=np.empty(0,dtype='float')  # confused about the use of 0
    nrg0=np.empty((1,ncols))
    nrg1=np.empty((0,ncols),dtype='float')
    
    #grab 'n_spec' from 'parameters'
#    nspec = get_nspec(out_dir,suffix)
    nspec = pars['n_spec']
    
    #separate initializations for different 'n_spec' values
    if nspec<=2:
        nrg2=np.empty((0,ncols),dtype='float')
    if nspec<=3:
        nrg2=np.empty((0,ncols),dtype='float')
        nrg3=np.empty((0,ncols),dtype='float')
    
    
    #open 'nrg' file
    f=open(os.path.join(out_dir , 'nrg' + suffix),'r')
    nrg_in=f.read()

    #format 'nrg' file for reading
    nrg_in_lines=nrg_in.split('\n')
    for j in range(len(nrg_in_lines)):
        if nrg_in_lines[j] and j % (nspec+1) == 0:
            time=np.append(time,nrg_in_lines[j])
        elif nrg_in_lines[j] and j % (nspec+1) == 1:
            nline=nrg_in_lines[j].split()
            for i in range(ncols):
                nrg0[0,i]=nline[i]
            nrg1=np.append(nrg1,nrg0,axis=0)
        elif nspec>=2 and nrg_in_lines[j] and j % (nspec+1) ==2:
            nline=nrg_in_lines[j].split()
            for i in range(ncols):
                nrg0[0,i]=nline[i]
            nrg2=np.append(nrg2,nrg0,axis=0)
        elif nspec==3 and nrg_in_lines[j] and j % (nspec+1) ==3:
            nline=nrg_in_lines[j].split()
            for i in range(ncols):
                nrg0[0,i]=nline[i]
            nrg3=np.append(nrg3,nrg0,axis=0)

    #return 'time' and 'nrgx' arrays
    if nspec==1:
        return time,nrg1
    elif nspec==2:
        return time,nrg1,nrg2
    else:
        return time,nrg1,nrg2,nrg3

    
def get_parsed_params(filepath):
    par = Parameters()
    par.Read_Pars(filepath)
    pars = par.pardict 
    
    return pars

def get_suffixes(out_dir, sim_type):
    suffixes = []

    if sim_type=='GENE': ## scan all file with parameters 
        suffixes = [ os.path.basename(file).split('parameters')[-1] for file in glob.glob(os.path.join(out_dir,'parameters*')) if os.path.isfile(file)]
    elif sim_type in ['CGYRO','TGLF','GS2','GX']:  ## scan folders return as list 
        # suffixes = [os.path.basename(fldr) for fldr in os.listdir(out_dir) os.path.isdir(os.path.join(out_dir,fldr))]
        suffixes = [fldr for fldr in os.listdir(out_dir) if os.path.isdir(os.path.join(out_dir,fldr))]
    suffixes.sort()
    return suffixes

def get_gyrokinetics_from_dir(out_dir,user, linear):
    suffixes = get_suffixes(out_dir)
    numscan = len(suffixes)
    assert numscan>0, "files must have a suffix!"
    suffixes.sort()
    GK_list = []
    for suffix in suffixes:
        GK_list.append(get_gyrokinetics_from_run(out_dir, suffix, user, linear))
    
    return GK_list

def get_gyrokinetics_from_run(out_dir, suffix, user, linear, tspan = None):
    if isinstance(tspan, list):
        starttime = tspan[0]
        endtime = tspan[-1]
    else:
        starttime = -1
        endtime=-2
    
    current_pwd = os.getcwd()
    os.chdir(out_dir) # change dir or make it as an argument
    common = comm.CommonData(suffix, starttime, endtime)
    rundatafiles = datafiles.RunDataFiles(common)
    if linear:
        tst = GKDB_linear(common, rundatafiles, user)
    else:
        tst = GKDB_nonlin(common, rundatafiles, user)
    
    os.chdir(current_pwd) # get back to original dir
    
    return tst.gkdict

def get_diag_from_run(out_dir, suffix, t_span = None):
#    t_start = 0.0 # use start/end time in nrg files?
#    t_end = 100.0

    par0 = Parameters()
    par0.Read_Pars(os.path.join(out_dir , 'parameters' + suffix))
    pars0 = par0.pardict 

    if pars0['nonlinear'] == 'T' or pars0['nonlinear'] == True:
        nonlinear = True
    else:
        nonlinear = False

    if t_span is not None and len(t_span)==2:
        t_start = t_span[0]
        t_end = t_span[-1]
        
    else:
        print('Time span not speficied, searching it in NRG file\n')
        
        nspec = pars0['n_spec']
        if os.path.isfile(os.path.join(out_dir , 'nrg' + suffix) ):
            if nspec==1:
                tn,_ =get_nrg(out_dir, suffix)
                t_start = float(tn[0])
                t_end = float(tn[-1])
            elif nspec==2:
                tn,_,_ =get_nrg(out_dir, suffix)
                t_start = float(tn[0])
                t_end = float(tn[-1])
            elif nspec==3:
                tn,_,_,_ =get_nrg(out_dir, suffix)
                t_start = float(tn[0])
                t_end = float(tn[-1])
            else:
                exit("Not ready for nspec>3")
        else:
            t_start = 0.0
            t_end = 100.0
            print('NRG files not found for suffix {}. Using t_start = 0.0 and t_end = 100.0'.format(suffix))
        

    print('****** Diagnostics start at {} and end at {}. ******'.format(t_start, t_end))
    
    Diag_dict = {}
       
    #all is also included in the run object
    simulation=Simulation(out_dir, None, [suffix])
    run = simulation.runs[0]

    data = Data(run.parameters,suffix)
    
    #selected_diags= {'Flux Spectra':DiagFluxSpectra(avail_vars=data.av_vars, specnames=run.specnames),
    #                 'Amplitude Spectra':DiagAmplitudeSpectra(avail_vars=data.av_vars, specnames=run.specnames,parameters = run.parameters[0].pardict,spatialgrid = run.spatialgrid[0]),
    
    '''
    Creat a folder for saving plots if it does not exist
    '''
    if nonlinear:   #Only calculate spectra for nonlinear
        selected_diags = []
        diag_keys = []
        #selected_diags.append(DiagFieldMomSnapshots())
        #diag_keys.append('Field Mom Snapshots')
        selected_diags.append(DiagFluxSpectra())
        diag_keys.append('Flux Spectra')
        selected_diags.append(DiagAmplitudeSpectra())
        diag_keys.append('Amplitude Spectra')

        loader=Loader()
        step = 1
        loader.set_interval(selected_diags, simulation.data,run, t_start,t_end,step)
        for it, time in enumerate(loader.times):
            print(" time {}".format(time))
            for i_d, diag in enumerate(selected_diags):
                diag.execute(simulation.data, simulation.runs[0], loader.steps[it], loader.files[it],time)
        for i in range(len(diag_keys)): 
            print(diag_keys[i])
            Diag_dict[diag_keys[i]] = selected_diags[i].dict_to_mgkdb()

    '''
    Get appropriate time information for snapshots.  We want maximum of 30, minumum of 15
    Note: the sparse z-grid snapshots are only done for nonlinear
    '''
    ntime = len(simulation.data.av_times['mom'].times)
    if nonlinear:
        step = int(np.floor(ntime/30)) + 1
    else:
        step = 1
    '''
    Get snapshots and final fields and moments
    '''
    if nonlinear:
        print("Getting snapshots and final for fields and moments")
    else:
        print("Getting final for fields and moments")
    #print("Number of time points, step",ntime,step)
    selected_diags = []
    diag_keys = []
    selected_diags.append(DiagFieldMomSnapshots())
    diag_keys.append('Field Mom Snapshots')
 
    loader2=Loader()
    loader2.set_interval(selected_diags, simulation.data,run, t_start,t_end,step)
    for it, time in enumerate(loader2.times):
        if nonlinear or time == loader2.times[-1]:  #Only final time for linear
            print(" time {}".format(time))
            for i_d, diag in enumerate(selected_diags):
                diag.execute(simulation.data, simulation.runs[0], loader2.steps[it], loader2.files[it],time)
    for i in range(len(diag_keys)): 
        print(diag_keys[i])
        Diag_dict[diag_keys[i]] = selected_diags[i].dict_to_mgkdb()
    
    '''
    Grid
    '''
    Diag_dict['Grid'] = vars(run.spatialgrid)

    return Diag_dict
    
        



""" Base module for plotting

Contains:
- Matplotib imports and settings

"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
from pkg_resources import parse_version

if parse_version(mpl.__version__) >= parse_version("1.5"):
    from cycler import cycler


class Plotting:
    """ Base class for plotting routines

    Contains the parameters used for the plots such as
    line colors, and font sizes
    Sets the line and point marker widths/sizes
    Sets the default color cycle (respecting an API change in matplotlib 1.5)
    """

    def __init__(self):
        # Matplotlib 2.x changes the default plot styles
        if parse_version(mpl.__version__) >= parse_version("2.0"):
            # Colormap for quantities >0 (e.g. heat flux)
            self.cmap_unidirect = mpl.cm.viridis
            mpl.rc('font', size=15)
        else:
            self.cmap_unidirect = mpl.cm.viridis
            mpl.rc('font', size=15)
        # for quantities with critical value (e.g. 0 for phi)
        self.cmap_bidirect = mpl.cm.bwr
        self.color_list = plt.cm.Dark2(np.linspace(0, 1.0, 9))  # Plot line colors
        # Set some basic plot properties
        # Those are set globally everytime a Plotting object is instantiated!
        mpl.rc('lines', linewidth=2, markeredgewidth=1.5, markersize=10)
        # This sets the figure frames to transparent both on screen and in calls to savefig
        mpl.rc('figure', facecolor=(1, 1, 1, 0))
        mpl.rc('figure', frameon=False)
        mpl.rc('savefig', facecolor=(1, 1, 1, 0))
        #mpl.rc('savefig', frameon=False)
        # mpl 1.5 introduces a new, more flexible prop_cycle parameter, so different line styles
        # can be defined in the else case as well
        if parse_version(mpl.__version__) <= parse_version("1.4"):
            mpl.rc('axes', color_cycle=self.color_list)
        else:
            mpl.rc('axes', prop_cycle=cycler('color', self.color_list))
        # Dictionary that converts internal names to well-formatted LaTeX for the plots
        self.titles = {'omts': r'$a/L_T$', 'omns': r'$a/L_n$', 'ns': r'$n$', 'Ts': r'$T$',
                       'Gammaturb': r'$\Gamma_{\mathrm{turb}}/\Gamma_{gB}$',
                       'Gammanc': r'$\Gamma_{\mathrm{NC}}/\Gamma_{gB}$',
                       'Qturb': r'$Q_{\mathrm{turb}}/Q_{gB}$', 'Qnc': r'$Q_{\mathrm{NC}}/Q_{gB}$',
                       'Piturb': r'$\Pi_{\mathrm{turb}}/\Pi_{gB}$',
                       'Pinc': r'$\Pi_{\mathrm{turb}}/\Pi_{gB}$', 
                       "jbs": r'$j_\mathrm{BS}$',
                       "phi": r"$\phi$", 
                       "apar": r"$A_\parallel$", 
                       "dens": r"$n$",
                       "tpar": r"$T_\parallel$", 
                       "tperp": r"$T_\perp$",
                       "qpar": r"$q_\parallel + 1.5p_0 u_\parallel$",
                       "qperp": r"$q_\perp + p_0 u_\parallel$", 
                       "upar": r"$u_\parallel$",
                       "densI1": r"$nI1$",
                       "TparI1": r"$TparI1$",
                       "TppI1": r"$TppI1$",
                       "bpar":  r"$B_\parallel$"}

    def plot_contour_quant(self, ax, fig, x, y, data, is_complex, xlabel=None, ylabel=None,
                           title=None, is_bidirect=False, axcb=None):
        """ Plot a 2d quantity as a color map

        :param ax: Axis object to plot into
        :param fig: Figure object ax belongs to (needed for colorbar)
        :param x: x-axis grid
        :param y: y-axis grid
        :param data: 2d-array in x-y order
        :param is_complex: Takes the absolute value of data before plotting (for complex arrays)
        :param xlabel: Label for x-axis, None will leave it unchanged
        :param ylabel: Label for y-axis, None will leave it unchanged
        :param title: Axis title, None will leave it unchanged
        :param is_bidirect: Use a bidirectional colormap for data
        :param axcb: Axes object of the colorbar, needed for time series
        :return: Axes object of the colorbar for reuse
        """
        mapdata = np.abs(np.squeeze(data)).T if is_complex else np.squeeze(data).T
        if is_bidirect:
            posval = np.abs(np.nanmax(mapdata))
            negval = np.abs(np.nanmin(mapdata))
            vmax = posval if posval > negval else negval
            vmin = -vmax
            cmap = self.cmap_bidirect
        else:
            vmax, vmin = None, None
            cmap = self.cmap_unidirect
        cm1 = ax.pcolormesh(x, y, mapdata, cmap=cmap, vmax=vmax, vmin=vmin)
        ax.set_rasterization_zorder(z=-10)
        if xlabel:
            ax.set_xlabel(xlabel)
        if ylabel:
            ax.set_ylabel(ylabel)
        if title:
            ax.set_title(title)
        axcb = fig.colorbar(cm1, cax=axcb).ax
        return axcb


    def plot_1d_quant(self, ax, fig, x, data, is_complex, xlabel=None, title=None, axcb=None):
        """ Plot a 1d quantity

        :param ax: Axis object to plot into
        :param fig: Figure object ax belongs to (needed for colorbar)
        :param x: x-axis grid
        :param data: 1d-arra
        :param is_complex: Takes the absolute value of data before plotting (for complex arrays)
        :param xlabel: Label for x-axis, None will leave it unchanged
        :param title: Axis title, None will leave it unchanged
        """

        ax.plot(x, np.absolute(data.T) if is_complex else data.T, color='b')
        ax.set_rasterization_zorder(z=-10)

        if xlabel:
            ax.set_xlabel(xlabel)
        if title:
            ax.set_title(title)
        axcb = ax.axes
        return axcb

    def plot_contour_quant_timeseries(self, ax, fig, x, y, data, is_complex, times, xlabel=None,
                                      ylabel=None, title=None, is_bidirect=False):
        """ Plot a 2d quantity as a navigatable series of color maps

        :param ax: Axis object to plot into
        :param fig: Figure object ax belongs to (needed for colorbar)
        :param x: x-axis grid
        :param y: y-axis grid
        :param data: 2d-array in x-y order
        :param is_complex: Takes the absolute value of data before plotting (for complex arrays)
        :param times:
        :param xlabel: Label for x-axis, None will leave it unchanged
        :param ylabel: Label for y-axis, None will leave it unchanged
        :param title: Axis title, None will leave it unchanged
        :param is_bidirect: Use a bidirectional colormap for data
        """
        def key_event(event):
            nonlocal icurr_time
            keyresult = self._keyhandler(event, icurr_time, len(times))
            if keyresult is None:
                return
            else:
                icurr_time = keyresult
            # Clear both the axes (leaving labels) and the colorbar object
            for artist in ax.lines+ax.collections:
                artist.remove()
            axcb.clear()
            self.plot_contour_quant(ax, fig, x, y, data[icurr_time], is_complex,
                                    title="{}, t={:.3f}".format(title, times[icurr_time]),
                                    is_bidirect=is_bidirect, axcb=axcb)
            fig.canvas.draw()

        icurr_time = 0
        axcb = self.plot_contour_quant(ax, fig, x, y, data[icurr_time], is_complex, xlabel, ylabel,
                                title="{}, t={:.3f}".format(title, times[icurr_time]),
                                is_bidirect=is_bidirect)
        fig.canvas.mpl_connect('key_press_event', key_event)




    def plot_1d_quant_timeseries(self, ax, fig, x, data, is_complex, times, xlabel=None, title=None):
        """ Plot a 2d quantity as a navigatable series of color maps

        :param ax: Axis object to plot into
        :param fig: Figure object ax belongs to (needed for colorbar)
        :param x: x-axis grid
        :param data: 1d-array
        :param is_complex: Takes the absolute value of data before plotting (for complex arrays)
        :param times:
        :param xlabel: Label for x-axis, None will leave it unchanged
        """
        def key_event(event):
            nonlocal icurr_time
            keyresult = self._keyhandler(event, icurr_time, len(times))
            if keyresult is None:
                return
            else:
                icurr_time = keyresult
            # Clear both the axes (leaving labels)
            for artist in ax.lines+ax.collections:
                artist.remove()
            axcb.clear()
            self.plot_1d_quant(ax, fig, x, data[icurr_time], is_complex, xlabel,
                                    title="{}, t={:.3f}".format(title, times[icurr_time]), axcb=axcb)
            fig.canvas.draw()

        icurr_time = 0
        axcb = self.plot_1d_quant(ax, fig, x, data[icurr_time], is_complex, xlabel,
                           title="{}, t={:.3f}".format(title, times[icurr_time]))
        fig.canvas.mpl_connect('key_press_event', key_event)

    @staticmethod
    def _keyhandler(event, icurr_time, timelen):
        new_ind = icurr_time
        if event.key == "pageup":
            new_ind += 10
        elif event.key == "pagedown":
            new_ind -= 10
        elif event.key == "right":
            new_ind += 1
        elif event.key == "left":
            new_ind -= 1

        elif event.key == "home":
            new_ind = 0
        elif event.key == "end":
            new_ind = timelen-1
        else:
            return None
        return new_ind%timelen

    @staticmethod
    def _minlogaxis(arr):
        """ Calculate lower limit for plotting on a log axis

        lower Integer * 10^minimumexponent

        :param arr: Array to base the limit on
        """
        minim = arr[arr != 0].min()  # ignore 0
        min_exp = np.floor(np.log10(minim))
        min_mant = np.floor(minim*10 ** (-min_exp))
        floormin = min_mant*10 ** min_exp
        return floormin

    @staticmethod
    def _maxlogaxis(arr):
        """ Calculate upper limit for plotting on a log axis

         higher Integer * 10^maximumexponent

        :param arr: Array to base the limit on
        """
        maxim = arr[arr != 0].max()
        max_exp = np.floor(np.log10(maxim))
        max_mant = np.ceil(maxim*10 ** (-max_exp))
        ceilmax = max_mant*10 ** max_exp
        return ceilmax

    @staticmethod
    def gyrobohm_SI(rundata, quantity):
        """ Convert gyroBohm unit to SI unit

        :param rundata: Run object
        :param quantity: Name of the quantity to calculate gB for
        """
        # pylint: disable=invalid-name
        pnt = rundata
        elementary_charge = 1.60217662e-19
        temp_unit = 1e3*elementary_charge  # temp is in keV
        dens_unit = 1e19  # m**(-3)
        mass_unit = 1.6726219e-27  # proton mass
        # Magnetic field and reference length are already in T and m
        nref = pnt.pars["Lref"]*dens_unit
        Tref = pnt.pars["Tref"]*temp_unit
        mref = pnt.pars["mref"]*mass_unit
        cref = np.sqrt(Tref/mref)
        rhoref = mref*cref/elementary_charge/pnt.pars["Bref"]
        Gamma_gB = nref*cref*rhoref ** 2/pnt.pars["Lref"] ** 2
        gBunits = {"Ts": Tref, "ns": nref, "omts": 1, "omns": 1, "Gammanc": Gamma_gB/1e19,
                   "Gammaturb": Gamma_gB/1e19, "Qturb": Gamma_gB*Tref/1e3, "Qnc": Gamma_gB*Tref/1e3,
                   "Pinc": Gamma_gB*cref*mref, "Piturb": Gamma_gB*cref*mref,
                   "jbs": pnt.pars["Bref"]*nref*cref*rhoref/pnt.pars["Lref"]}
        return gBunits[quantity]



# -*- coding: utf-8 -*-
"""
Created on Mon Feb 17 16:05:27 2020

@author: dykua

A Plotting class for visualizing the diagnostic plots from mgk_fusion
"""
import numpy as np
#import matplotlib
#matplotlib.use('GTK') # uncomment if using MAC
import scipy.linalg as linalg
import matplotlib.pyplot as plt
from .putils import averages
from .diagnostics.baseplot import Plotting
import time
import os

class diag_plot():
    def __init__(self, data_dict, save_fig = False, save_dir = './'):
        '''
        data_dict retrieved from database via 'load_diag' method
        '''
        self.data = data_dict['Diagnostics']
        self._id = data_dict['_id']
        self.meta = data_dict['Metadata'] if 'Metadata' in data_dict else None
        self.save_fig = save_fig
        self.save_dir = save_dir

        self.avail_plts = {"Amplitude Spectra": self.diag_amplitude_spectra,
                           "Flux Spectra": self.diag_flux_spectra,
                           "Shearing Rate": self.diag_shearing_rate,
                           "Freq Growth Rate": self.diag_freqgrowrate,
                           "Contours": self.diag_contours,
                           #"Ballamp": self.diag_ballamp,
                           "Cross Phase": self.diag_crossphase}
        
        self.avail_diags = self.avail_plts.keys()
        
    def diag_amplitude_spectra(self):
        
        kx = self.data['Grid']['kx_pos']
        ky = self.data['Grid']['ky']
        
        time_requested = self.data['Time']
        
        plotbase = Plotting()
        plotbase.titles.update(
                {"Ges": r"$\Gamma_{es}$", "Qes": r"$Q_{es}$", "Pes": r"$\Pi_{es}$",
                 "Gem": r"$\Gamma_{em}$", "Qem": r"$Q_{em}$", "Pem": r"$\Pi_{em}$"})
    
        for quant in self.data['Amplitude Spectra']['kx'].keys():

            fig = plt.figure(figsize=(6, 8))
#            fig_list.append(fig)

            ax_loglog_kx = fig.add_subplot(3, 2, 1)
            ax_loglin_kx = fig.add_subplot(3, 2, 3)
            ax_linlin_kx = fig.add_subplot(3, 2, 5)
            ax_loglog_ky = fig.add_subplot(3, 2, 2)
            ax_loglin_ky = fig.add_subplot(3, 2, 4)
            ax_linlin_ky = fig.add_subplot(3, 2, 6)

            amplitude_kx = averages.mytrapz(self.data['Amplitude Spectra']['kx'][quant], time_requested)
            amplitude_ky = averages.mytrapz(self.data['Amplitude Spectra']['ky'][quant], time_requested)

            # log-log plots, dashed lines for negative values
            baselogkx, = ax_loglog_kx.plot(kx, amplitude_kx)
            baselogky, = ax_loglog_ky.plot(ky, amplitude_ky)

            # lin-log plots, nothing fancy
            baseloglinkx, = ax_loglin_kx.plot(kx, amplitude_kx)
            baseloglinky, = ax_loglin_ky.plot(ky, amplitude_ky)

            # lin-lin plots, nothing fancy
            ax_linlin_kx.plot(kx, amplitude_kx)
            ax_linlin_ky.plot(ky, amplitude_ky)

            # set things
            ax_loglog_kx.loglog()
            ax_loglog_ky.loglog()

            ax_loglin_kx.set_xscale("log")
            ax_loglin_ky.set_xscale("log")

            # lin-lin plots, nothing fancy
            ax_linlin_kx.set_xlim(left=0)
            ax_linlin_kx.set_xlabel(r"$k_x \rho_{ref}$")
            ax_linlin_ky.set_xlabel(r"$k_y \rho_{ref}$")

            ind = quant.find('#')
            if ind == -1:
                ax_loglog_ky.set_title("{}".format(quant))
                ax_loglog_kx.set_title("{}".format(quant))
            else:
                ax_loglog_ky.set_title("{}".format(quant[0:ind] + " " + quant[ind + 1:]))
                ax_loglog_kx.set_title("{}".format(quant[0:ind] + " " + quant[ind + 1:]))
                
#            if self.meta is not None:
#                fig.suptitle(str(self._id) + ' from ' + str(self.meta))
#            else:
            fig.suptitle( str(self._id) )
                
            fig.tight_layout()
            #plt.show()
            fig.show() 
            #if self.save_fig:
            #    fig.savefig(os.path.join(self.save_dir, 'AmplitudeSpectra-{}-{}.png'.format(quant, time.strftime("%y%m%d-%H%M%S"))) )

    def diag_flux_spectra(self):
        
        kx = self.data['Grid']['kx_pos']
        ky = self.data['Grid']['ky']
        
        time_requested = self.data['Time']
        
        plotbase = Plotting()
        plotbase.titles={"Ges": r"$\Gamma_{es}$", "Qes": r"$Q_{es}$", "Pes": r"$\Pi_{es}$"}
#        if self.has_EM:
        plotbase.titles.update({"Gem": r"$\Gamma_{em}$", "Qem": r"$Q_{em}$", "Pem": r"$\Pi_{em}$"})
        
        
        for spec in self.data['Flux Spectra'].keys():
            fig = plt.figure(figsize=(6, 8))

            ax_loglog_kx = fig.add_subplot(3, 2, 1)
            ax_loglin_kx = fig.add_subplot(3, 2, 3)
            ax_linlin_kx = fig.add_subplot(3, 2, 5)
            ax_loglog_ky = fig.add_subplot(3, 2, 2)
            ax_loglin_ky = fig.add_subplot(3, 2, 4)
            ax_linlin_ky = fig.add_subplot(3, 2, 6)

            spec_flux = self.data['Flux Spectra'][spec]
            
#            if self.meta is not None:
#                fig.suptitle(str(self._id) + ' from ' + str(self.meta))
#            else:
            fig.suptitle( str(self._id) )

            for flux in spec_flux.keys():

                flux_kx = averages.mytrapz(spec_flux[flux]['kx'], time_requested)
                flux_ky = averages.mytrapz(spec_flux[flux]['ky'], time_requested)
                # Mask negative flux values for solid lines

                pos_flux_kx = np.ma.masked_where((flux_kx <= 0), flux_kx)
                # Mask zero flux for dashed lines, this takes care of the Nyquist mode in kx
                all_flux_kx = np.ma.masked_where((flux_kx == 0), flux_kx)

                pos_flux_ky = np.ma.masked_where((flux_ky <= 0), flux_ky)
                all_flux_ky = np.ma.masked_where((flux_ky == 0), flux_ky)

                # log-log =plots, dashed lines for negative values
                baselogkx, = ax_loglog_kx.plot(kx, pos_flux_kx, label=plotbase.titles[flux])
                ax_loglog_kx.plot(kx, np.abs(all_flux_kx), ls="--", color=baselogkx.get_color())
                baselogky, = ax_loglog_ky.plot(ky, pos_flux_ky, label=plotbase.titles[flux])
                ax_loglog_ky.plot(ky, np.abs(all_flux_ky), ls="--", color=baselogky.get_color())

                # lin-log plots, nothing fancy
                baseloglinkx, = ax_loglin_kx.plot(kx, all_flux_kx, label=plotbase.titles[flux])
                ax_loglin_kx.plot(kx, all_flux_kx*kx, ls="--", color=baseloglinkx.get_color())
                baseloglinky, = ax_loglin_ky.plot(ky, all_flux_ky, label=plotbase.titles[flux])
                ax_loglin_ky.plot(ky, all_flux_ky*ky, ls="--", color=baseloglinky.get_color())

                # lin-lin plots, nothing fancy
                ax_linlin_kx.plot(kx, all_flux_kx, label=plotbase.titles[flux])
                ax_linlin_ky.plot(ky, all_flux_ky, label=plotbase.titles[flux])

#                str_out = "{} {} = {:.4f} (kx intg.) - {:.4f} (ky intg.)".format(spec, flux,
#                                                                                 np.sum(flux_kx),
#                                                                                 np.sum(flux_ky))

            # set things
            ax_loglog_kx.loglog()
            ax_loglog_kx.set_xlabel(r"$k_x \rho_{ref}$")

            ax_loglog_ky.loglog()
            ax_loglog_ky.set_xlabel(r"$k_y \rho_{ref}$")

            ax_loglin_kx.set_xscale("log")
            ax_loglin_ky.set_xscale("log")

            # lin-lin plots, nothing fancy
            ax_linlin_kx.set_xlim(left=0)
            ax_linlin_kx.set_xlabel(r"$k_x \rho_{ref}$")

            ax_linlin_ky.set_xlabel(r"$k_y \rho_{ref}$")

            for ax in [ax_loglog_kx, ax_loglin_kx, ax_linlin_kx, ax_loglog_ky, ax_loglin_ky,
                       ax_linlin_ky, ]:
                # ax.set_ylabel(r"$<|A|^2>$")
                ax.legend()
            ax_loglog_ky.set_title("{}".format(spec))
            ax_loglog_kx.set_title("{}".format(spec))
            #            fig.tight_layout()
            #plt.show()
            fig.show()
            #if self.save_fig:
            #    fig.savefig(os.path.join(self.save_dir, 'FluxSpectra-{}-{}-{}.png'.format(spec,flux, time.strftime("%y%m%d-%H%M%S"))))
    
    def diag_shearing_rate(self):
        
        x = self.data['Grid']['x']
#        ky = self.data['Grid']['ky']
        
        time_requested = self.data['Time']        
#        plotbase = Plotting()
        
        def plot_a_map(ax, x, y, f, x_lbl, y_lbl, ttl, plotbase=Plotting()):
            cm1 = ax.pcolormesh(x, y, f)
#            cm1 = ax.contourf(x, y, f, 100, cmap=plotbase.cmap_bidirect)
            ax.set_rasterization_zorder(z=-10)
            ax.set_xlabel(x_lbl)
            ax.set_ylabel(y_lbl)
            ax.set_title(ttl)
            fig.colorbar(cm1)

        x_lbl = r'$x/\rho_{ref}$' if self.data['Shearing Rate']['x_local'] else r'x/a'

        if len(time_requested) > 1:
            # some maps
            fig = plt.figure()
#            if self.meta is not None:
#                fig.suptitle(str(self._id) + ' from ' + str(self.meta))
#            else:
            fig.suptitle( str(self._id) )
                
            ax = fig.add_subplot(2, 2, 1)
            plot_a_map(ax, time_requested, x,
                       self.data['Shearing Rate']['phi_zonal_x'].T,
                       r'$ t c_{ref}/L_{ref}$ ', x_lbl, r'$ \langle\phi\rangle [c_{ref}/L_{ref}]$')

            ax = fig.add_subplot(2, 2, 2)
            plot_a_map(ax, time_requested, x,
                       self.data['Shearing Rate']['Er_x'].T, r'$t c_{ref}/L_{ref}$',
                       x_lbl, r'$E_r [eT_{ref}/ (\rho^*_{ref})^2 L_{ref}]$')

            ax = fig.add_subplot(2, 2, 3)
            plot_a_map(ax, time_requested, x,
                       self.data['Shearing Rate']['vExB_x'].T, r'$t c_{ref}/L_{ref}$',
                       x_lbl, r'$v_{ExB} [c_{ref} \rho^*_{ref}]$')

            ax = fig.add_subplot(2, 2, 4)
            plot_a_map(ax, time_requested, x,
                       self.data['Shearing Rate']['omegaExB_x'].T,
                       r'$t c_{ref}/L_{ref}$', x_lbl, r'$\omega_{ExB} [c_{ref}/L_{ref}]$')
            fig.show()
            #if self.save_fig:
            #    fig.savefig(os.path.join(self.save_dir, 'ShearingRate-map-{}.png'.format(time.strftime("%y%m%d-%H%M%S"))))
            #plt.show()

            # time traces
            my_pos = self.data['Shearing Rate']['my_pos']
#            print(my_pos)
            fig = plt.figure()
#            if self.meta is not None:
#                fig.suptitle(str(self._id) + ' from ' + str(self.meta))
#            else:
            fig.suptitle( str(self._id) )
            ax = fig.add_subplot(2 + self.data['Shearing Rate']['x_local'], 1, 1)
#            print(self.data['Shearing Rate']['vExB_x'][my_pos].shape)
            ax.plot(time_requested, self.data['Shearing Rate']['vExB_x'][:,my_pos].T)
            ax.set_xlabel(r'$t c_{ref}/L_{ref}$')
            ax.set_ylabel(r'$v_{ExB} [c_{ref} \rho^*_{ref}]$')

            ax = fig.add_subplot(2 + self.data['Shearing Rate']['x_local'], 1, 2)
            ax.plot(time_requested,
                    self.data['Shearing Rate']['omegaExB_x'][:,my_pos].T)
            ax.set_xlabel(r'$t c_{ref}/L_{ref}$')
            ax.set_ylabel(r'$\omega_{ExB} [c_{ref} \rho^*_{ref}]$')

            if self.data['Shearing Rate']['x_local']:
                ax = fig.add_subplot(2 + self.data['Shearing Rate']['x_local'], 1, 3)
                ax.plot(time_requested, np.sqrt(np.mean(np.power(np.abs(self.data['Shearing Rate']['omegaExB_x']), 2),axis=1)).T) 
                         
                ax.set_xlabel(r'$t c_{ref}/L_{ref}$')
                ax.set_ylabel(r'$\sqrt{|\omega_{ExB}|^2} [c_{ref}/L_{ref}]$')

#                shear_avg = averages.mytrapz(np.array(
#                        [np.sqrt(np.mean(np.power(np.abs(x.omegaExB_x), 2))) for x in
#                         self.shearing_rate]), time_requested)
#                str_out = "ExB shearing rate= {:.3f}".format(shear_avg)
#                if output:
#                    output.info_txt.insert(END, str_out + "\n")
#                    output.info_txt.see(END)

            fig.show()
            #if self.save_fig:
            #    fig.savefig(os.path.join(self.save_dir, 'ShearingRate-TT.png'.format(time.strftime("%y%m%d-%H%M%S")) ))
            #plt.show()

        # zonal spectra
        if self.data['Shearing Rate']['x_local']:
            fig = plt.figure()
#            if self.meta is not None:
#                fig.suptitle(str(self._id) + ' from ' + str(self.meta))
#            else:
            fig.suptitle( str(self._id) )
            ax = fig.add_subplot(3, 1, 1)
            ax.plot(self.data['Grid']['kx_pos'], averages.mytrapz(self.data['Shearing Rate']['abs_phi_fs'], time_requested))
            ax.set_xlabel(r'$k_x \rho_{ref}$')
            ax.set_title(r"$|\langle\phi\rangle|$")
            ax.loglog()

            ax = fig.add_subplot(3, 1, 2)
            ax.plot(self.data['Grid']['kx_pos'], averages.mytrapz(self.data['Shearing Rate']['abs_phi_fs'], time_requested))
            ax.set_xlabel(r'$k_x \rho_{ref}$')
            ax.set_xscale("log")

            ax = fig.add_subplot(3, 1, 3)
            ax.plot(self.data['Grid']['kx_pos'], averages.mytrapz(self.data['Shearing Rate']['abs_phi_fs'], time_requested))
            ax.set_xlabel(r'$k_x \rho_{ref}$')

            fig.show()
            #if self.save_fig:
            #    fig.savefig(os.path.join(self.save_dir, 'ShearingRate-ZS-{}.png'.format(time.strftime("%y%m%d-%H%M%S"))))
            #plt.show()

        # radial plots
        fig = plt.figure()
#        if self.meta is not None:
#            fig.suptitle(str(self._id) + ' from ' + str(self.meta))
#        else:
        fig.suptitle( str(self._id) )
        
        ax = fig.add_subplot(3, 1, 1)
        ax.plot(x,
                averages.mytrapz(self.data['Shearing Rate']['phi_zonal_x'],
                                 time_requested))
        ax.set_xlabel(x_lbl)
        ax.set_ylabel(r'$v_{ExB} [c_{ref} \rho^*_{ref}]$')

        ax = fig.add_subplot(3, 1, 2)
        ax.plot(x,
                averages.mytrapz(self.data['Shearing Rate']['vExB_x'], time_requested))
        ax.set_xlabel(x_lbl)
        ax.set_ylabel(r'$v_{ExB} [c_{ref} \rho^*_{ref}]$')

        ax = fig.add_subplot(3, 1, 3)
        ax.plot(x,
                averages.mytrapz(self.data['Shearing Rate']['omegaExB_x'],
                                 time_requested))
        ax.set_xlabel(x_lbl)
        ax.set_ylabel(r'$\omega_{ExB} [c_{ref} \rho^*_{ref}]$')

        fig.tight_layout()    
        #plt.show()
        fig.show()
        #if self.save_fig:
        #    fig.savefig(os.path.join(self.save_dir, 'ShearingRate-R-{}.png'.format(time.strftime("%y%m%d-%H%M%S"))))
    
#    def diag_ballamp(self):
#        plotbase = Plotting()
#        plotbase.titles.update(
#                {"phi": r"$\phi$", "A_par": r"$A_{\parallel}$", "B_par": r"$B_{\parallel}$",
#                 "dens": r"$n$", "T_par": r"T_{\parallel}", "T_perp": r"T_{\perp}",
#                 "u_par": r"u_{\parallel}", "q_par": r"q_{\parallel}",
#                 "q_perp": r"q_{\perp}"})
#
#        def plot_a_quant(fig, ncols, i_c, x, y, ttl, quant):
#
#            # lin-lin plots
#            ax_lin = fig.add_subplot(2, ncols, 1 + 2*(i_c - 1))
#            ax_lin.plot(x, np.absolute(y), label="|{}|".format(plotbase.titles[quant]),
#                        color='black')
#            if self.data['Ballamp']['opts']['re_im']['value']:
#                ax_lin.plot(x, np.real(y), label="Re({})".format(plotbase.titles[quant]),
#                            color='blue', lw=1)
#                ax_lin.plot(x, np.imag(y), label="Im({})".format(plotbase.titles[quant]),
#                            color='red', lw=1)
#            ax_lin.set_xlabel(r"$\chi/\pi$")
#            ax_lin.set_title(ttl)
#            ax_lin.axhline(y=0, color='black', lw=1, ls='--')
#            ax_lin.legend()
#
#            # log-log plots
#            ax_log = fig.add_subplot(2, ncols, 2 + 2*(i_c - 1))
#            ax_log.plot(x, np.absolute(y), label="|{}|".format(plotbase.titles[quant]),
#                        color='black')
#            if self.data['Ballamp']['opts']['re_im']['value']:
#                ax_log.plot(x, np.real(y), label="Re({})".format(plotbase.titles[quant]),
#                            color='blue', lw=1)
#                ax_log.plot(x, np.imag(y), label="Im({})".format(plotbase.titles[quant]),
#                            color='red', lw=1)
#            ax_log.set_xlabel(r"$\chi/\pi$")
#            ax_log.axhline(y=0, color='black', lw=1, ls='--')
#            ax_log.set_yscale('log')
#            ax_log.legend()
#
#        chi_pi = self.data['Ballamp']['chi_pi']
#        kyval = self.data['Ballamp']['kyval']
#        ttl_ky = r"$k_y\rho={:6.3f} $".format(kyval)
#
#        if self.data['Ballamp']['opts']['t_avg']['value']:
#            # loop over file
#            if self.data['Ballamp']['only field']:
#                nc = len(self.data['Ballamp']['ballamps'].keys() )
#                fig = plt.figure()
#                for i_q, quant in enumerate(self.data['Ballamp']['ballamps'].keys()):
#                    plot_a_quant(fig, nc, i_q + 1, chi_pi,
#                                 averages.mytrapz(self.data['Ballamp']['ballamps'][quant], self.data['Time']  ), ttl_ky,
#                                 quant)
#                    fig.suptitle( str(self._id) )
#                    fig.show()
#            else:
#                for quant in self.data['Ballamp']['ballamps'].keys():
#                    fig = plt.figure()
#                    #                    ttl = self.plotbase.titles[quant] if quant.find('#')==-1
#                    #                    else self.plotbase.titles[quant[0:quant.find(
#                    #                    '#')==-1]]+" "+quant[quant.find('#')==-1+1:]
#                    if quant.find('#') == -1:
#                        ttl = plotbase.titles[quant]
#                    else:
#                        ttl = quant[quant.find('#') + 1:]
#                    plot_a_quant(fig, 1, 1, chi_pi,
#                                 averages.mytrapz(self.data['Ballamp']['ballamps'][quant], self.data['Time'] ), ttl,
#                                 quant[0:quant.find('#')])
#                    fig.suptitle( str(self._id) )
#                    fig.show()
#
#        elif len(self.data['Time']) < 12:
#            # in this case we plot on a single figure for each quantity
#            if self.data['Ballamp']['only field']:
#                nc = len(self.data['Ballamp']['ballamps'].keys() )
#                for i_t, time in enumerate(self.data['Time']):
#                    fig = plt.figure()
#                    for i_q, quant in enumerate(self.data['Ballamp']['ballamps'].keys()):
#                        plot_a_quant(fig, nc, i_q + 1, chi_pi, self.data['Ballamp']['ballamps'][quant][i_t].T,
#                                     ttl_ky + "@ t={:.3f}".format(time), quant)
#                    fig.suptitle( str(self._id) )
#                    fig.show()
#            else:
#                for quant in self.data['Ballamp']['ballamps'].keys():
#                    ind = quant.find('#')
#                    if ind == -1:
#                        ttl = plotbase.titles[quant]
#                        leg = quant
#                    else:
#                        leg = quant[0:ind]
#                        ttl = quant[ind + 1:] + " "
#                    for i_t, time in enumerate(self.data['Time']):
#                        fig = plt.figure()
#                        plot_a_quant(fig, 1, 1, chi_pi, self.data['Ballamp']['ballamps'][quant][i_t].T,
#                                     ttl + ttl_ky + "@ t={:.3f}".format(time), leg)
#                        fig.suptitle( str(self._id) )
#                        fig.show()
    
    def diag_freqgrowrate(self):
        Freq_arr = []
        Grate_arr = []
        ky_arr = []

        def extract_freq(phi_pt, times_req, act_ky, plot_all, output=None):
            #   Amplitude and phase
            freq_trace = np.imag(np.diff(np.log(phi_pt)) / np.diff(times_req))
            gamma_trace = np.real(np.diff(np.log(phi_pt)) / np.diff(times_req))

            ph_f = np.angle(phi_pt)
            ampl_f = np.abs(phi_pt)

            #   linear regression of phase
            M = np.column_stack((np.ones(times_req.size), times_req))

            coeffsFreq = linalg.lstsq(M, ph_f)[0]
            coeffsAmp = linalg.lstsq(M, np.log(ampl_f))[0]

            Freq = coeffsFreq[1]
            Grate = coeffsAmp[1]

            Freq_arr.append(Freq)
            Grate_arr.append(Grate)

            if output:
                str_out = "ky mode {}: (Frequency, GrowthRate) {} {}\n".format(act_ky, Freq, Grate)
                output.info_txt.insert(END, str_out)

            if plot_all:
                ttl = r"$k_y\rho={:6.3f} $".format(act_ky)

                fig = plt.figure()

                ax_1 = fig.add_subplot(3, 1, 1)
                ax_1.plot(times_req, ampl_f, ls="-", color='b')
                ax_1.plot(times_req, np.exp(coeffsAmp[0] +
                                            coeffsAmp[1] * times_req),
                          ls="-", color='r')
                ax_1.set_xlabel(r"$t  c_{ref} / L_{ref}$")
                ax_1.set_yscale("log")
                ax_1.set_ylabel(r'$|\phi|$')
                ax_1.set_title(ttl)
                ax_1.text(0.05, 0.8, r'$\gamma={:4.3f}'.format(Grate) +
                          ' c_{ref}/L_{ref}$', transform=ax_1.transAxes)

                ax_2 = fig.add_subplot(3, 1, 2)
                ax_2.plot(times_req, ph_f, ls="-", color='b')
                ax_2.plot(times_req, coeffsFreq[0] + coeffsFreq[1] * times_req,
                          ls="--", color='r')
                ax_2.set_xlabel(r"$t  c_{ref} / L_{ref}$")
                ax_2.set_ylabel(r'$arg(\phi)$')
                ax_2.text(0.05, 0.8, r'$\omega={:4.3f}'.format(Freq) +
                          ' c_{ref}/L_{ref}$', transform=ax_2.transAxes)

                ax_3 = fig.add_subplot(3, 1, 3)
                ax_3.plot(times_req, np.abs(np.real(phi_pt)), ls='-',
                          color='b', label=r"$Re(|\phi|)$")
                ax_3.plot(times_req, np.abs(np.imag(phi_pt)), ls='--',
                          color='r', label=r"$Imag(|\phi|)$")
                ax_3.set_xlabel(r"$t  c_{ref} / L_{ref}$")
                ax_3.set_yscale("log")
                ax_3.legend()
                
                fig.suptitle( str(self._id) )
                fig.show()

                fig_t = plt.figure()
                a_1 = fig_t.add_subplot(2, 1, 1)
                a_1.plot(times_req[:-1], freq_trace, ls="-", color='b')
                a_1.set_xlabel(r"$t c_{ref}/L_{ref}$")
                a_1.set_ylabel(r"$\omega_r c_{ref}/L_{ref}$")
                a_1.set_title(ttl)

                a_2 = fig_t.add_subplot(2, 1, 2)
                a_2.plot(times_req[:-1], gamma_trace, ls="-", color='b')
                a_2.set_xlabel(r"$t c_{ref}/L_{ref}$")
                a_2.set_ylabel(r"$\gamma c_{ref}/L_{ref}$")
                
                fig_t.suptitle( str(self._id) )
                fig_t.show()

        plotbase = Plotting()
        plot_all = 1 if self.data['Freq Growth Rate']['opts']['show']['value'] else 0
        for ky in range(0, len(self.data['Freq Growth Rate']['phi'][0][:])):
            if (np.absolute(self.data['Freq Growth Rate']['ky'][ky]) <= np.absolute(self.data['Freq Growth Rate']['ky'][self.data['Freq Growth Rate']['kyind_max']])
                and np.absolute(self.data['Freq Growth Rate']['ky'][ky]) >= np.absolute(self.data['Freq Growth Rate']['ky'][self.data['Freq Growth Rate']['kyind_min']])):

                phi_trace = np.array([x[ky] for x in self.data['Freq Growth Rate']['phi']])
                act_ky = self.data['Freq Growth Rate']['ky'][ky]
                extract_freq(phi_trace, self.data['Time'], act_ky, plot_all, output=None)
                ky_arr.append(act_ky)

        fig_f = plt.figure()
        a_1 = fig_f.add_subplot(2, 1, 1)
        a_1.plot(ky_arr, Grate_arr, 'bo')
        a_1.set_xlabel(r"$k_y$")
        a_1.set_ylabel(r"$\gamma c_{ref}/L_{ref}$")
        a_2 = fig_f.add_subplot(2, 1, 2)
        a_2.plot(ky_arr, Freq_arr, 'bo')
        a_2.set_xlabel(r"$k_y$")
        a_2.set_ylabel(r"$\omega c_{ref}/L_{ref}$")
        
        fig_f.suptitle( str(self._id) )
        fig_f.show()
        
        
    def diag_crossphase(self):
        nbins = 64

        out = np.zeros((len(self.data['Grid']['ky']), nbins))

#        plotbase = Plotting()

        for spec in self.data['Cross Phase']['cross_phase'].keys():
            array = np.array(self.data['Cross Phase']['cross_phase'][spec])           
            for i_ky in range(len(self.data['Grid']['ky'])):
                phase_ky = array[:,:,i_ky,...].flatten()
#                phase_ky = self.cross_phase[spec][:][:][i_ky][:][:].flatten()
                hist, bin_edges = np.histogram(phase_ky/self.data['Cross Phase']['nx0']/self.data['Cross Phase']['nz0'],
                                               bins=nbins, range=(-np.pi, np.pi))
                out[i_ky, :] = hist

            fig = plt.figure()
            plt.xlabel('Phase angle/rad', fontsize=14)
            plt.ylabel(r'$k_y\rho_s$', fontsize=16)
            # cm1 = plt.contourf(bin_edges[1:],ky[1:],out[1:,:],50,cmap=self.plotbase.cmap_bidirect)
            cm1 = plt.pcolormesh(bin_edges[1:], self.data['Grid']['ky'][1:], out[1:, :])
            fig.colorbar(cm1)
            fig.tight_layout()
            fig.suptitle( str(self._id) )
            fig.show()
            
    def diag_contours(self):
        plotbase = Plotting()
        if self.data['Contours']['is3d']:
            plotbase.titles.update(
                    {"phi": r"$\phi$", "n": r"$n$", "u_par": r"u_{//}", "T_par": r"T_{//}",
                     "T_perp": r"T_{\perp}", "Q_es": r"$Q_{es}$", "Q_em": r"$Q_{em}$",
                     "Gamma_es": r"$\Gamma_{es}$", "Gamma_em": r"$\Gamma_{em}$"})
        else:
            plotbase.titles.update(
                    {"phi": r"$\phi$", "A_par": r"$A_{//}$", "B_par": r"$B_{//}$", "dens": r"$n$",
                     "T_par": r"T_{//}", "T_perp": r"T_{\perp}", "u_par": r"u_{//}",
                     "q_par": r"q_{//}", "q_perp": r"q_{\perp}"})
        
        if self.data['Contours']['opts']['f_x']['value']:
            x_lbl = r"$k_x\rho_{ref}$"
            x = self.data['Grid']['kx_fftorder']
        else:
            x_lbl = r"$x/\rho_{ref}$"
            x = self.data['Grid']['x']

        if self.data['Contours']['opts']['f_y']['value']:
            y_lbl = r"$k_y\rho_{ref}$"
            y = self.data['Grid']['ky']
        else:
            y_lbl = r"$y/\rho_{ref}$"
            y = self.data['Grid']['y']

        is_f = self.data['Contours']['opts']['f_x']['value'] or self.data['Contours']['opts']['f_y']['value']

        if self.data['Contours']['opts']['t_avg']['value']:
            # loop over file
            for quant in self.data['Contours']['contours'].keys():
                ind_str = quant.find('#')
                if ind_str == -1:
                    ttl = plotbase.titles[quant]
                else:
                    ttl = plotbase.titles[quant[0:ind_str]] + " " + quant[ind_str + 1:]

                fig = plt.figure()
                ax = fig.add_subplot(111)
                plotbase.plot_contour_quant(ax, fig, x, y,
                                            averages.mytrapz(self.data['Contours']['contours'][quant], self.data['Time']),
                                            is_f, x_lbl, y_lbl, ttl)
                fig.suptitle( str(self._id) )
                fig.show()
        else:
            for quant in self.data['Contours']['contours'].keys():
                ind_str = quant.find('#')
                if ind_str == -1:
                    ttl = plotbase.titles[quant]
                else:
                    ttl = plotbase.titles[quant[0:ind_str]] + " " + quant[ind_str + 1:]

                fig = plt.figure()
                ax = fig.add_subplot(111)
                plotbase.plot_contour_quant_timeseries(ax, fig, x, y, self.data['Contours']['contours'][quant], is_f,
                                                       self.data['Time'], x_lbl, y_lbl, ttl)
                fig.suptitle( str(self._id) )
                fig.show()
#        plt.show()
       
                
    def plot_all(self):
        for p in self.avail_plts.keys():
            self.avail_plts[p]()
    
    def print_avail_diags(self):
        print("Available diagnostics are: {}".format(self.avail_diags) )
        
    def plot_diag(self, key):
        if key in self.avail_plts.keys():
            self.avail_plts[key]()
        else:
            print("The specified diagnostic is not supported.")
            
#        self.diag_amplitude_spectra()
#        self.diag_flux_spectra()
#        self.diag_shearing_rate()
#        self.diag_contours()
#        self.diag_ballamp()
#        self.diag_crossphase()
#        self.diag_freqgrowrate()
        
    
        
        
        



""" Base classes for file reading"""

import struct
from bisect import bisect_left, bisect_right
import mmap

import numpy as np

from ..utils.averages import mytrapz


class BinaryFile(object):
    """ Base class to read Fortran binary (unformatted) files from GENE runs"""

    def __init__(self, filename, common, usemmap=False):
        self.cm = common
        self.file = None
        self.timearray = []
        self.tind = 0
        self.time = None
        self.filename = filename
        if usemmap:
            self.readfunc = np.frombuffer
        else:
            self.readfunc = self._npfromfilewithoffset
        self.usemmap = usemmap
        self._set_datatypes()
        self.te, self.tesize = self.time_entry()
        self._set_sizes()
        self.redirect(filename)  # Initialises self.file

    @staticmethod
    def _npfromfilewithoffset(file, count, dtype, offset):
        file.seek(offset)
        return np.fromfile(file, count=count, dtype=dtype)

    def redirect(self, filename):
        """ Call this routine to read from a new file"""
        self.filename = filename
        try:
            self.file.close()
        except (AttributeError, OSError):
            pass
        file = open(filename, 'rb')
        if self.usemmap:
            self.file = mmap.mmap(file.fileno(), length=0, access=mmap.ACCESS_READ)
        else:
            self.file = file
        self.timearray = []
        self.get_timearray()
        self.reset_tinds()

    def _set_datatypes(self):
        try:
            self.bigendian = self.cm.pars['ENDIANNESS'] == 'BIG'
        except KeyError:
            self.bigendian = False
        if self.bigendian:
            self.nprt = (np.dtype(np.float64)).newbyteorder()
            self.npct = (np.dtype(np.complex128)).newbyteorder()
        else:
            self.nprt = np.dtype(np.float64)
            self.npct = np.dtype(np.complex128)

    def _set_sizes(self):
        self.intsize = 4
        try:
            self.realsize = 8 if self.cm.pars['PRECISION'] == 'DOUBLE' else 4
        except KeyError:
            self.realsize = 8
        self.complexsize = 2*self.realsize

    def time_entry(self):
        """ Defines the struct for a time entry """
        # Fortran writes records as sizeof(entry) entry sizeof(entry)
        if self.bigendian:
            timeentry = struct.Struct('>idi')
        else:
            timeentry = struct.Struct('=idi')
        return timeentry, timeentry.size

    def _find_nearest_time(self, time):
        pos = bisect_left(self.timearray, time)
        if pos == 0:
            return self.timearray[0]
        if pos == len(self.timearray):
            return self.timearray[-1]
        before = self.timearray[pos - 1]
        after = self.timearray[pos]
        if after - time < time - before:
            return after
        else:
            return before

    def set_approximate_time(self, time):
        ti = self._find_nearest_time(time)
        self.set_time(ti)

    def set_time(self, time):
        """ Set current timestep by value"""
        self.time = time
        self.tind = self.timearray.index(time)

    def set_tind(self, tind):
        """ Set current timestep by index"""
        self.tind = tind
        self.time = self.timearray[tind]

    def get_minmaxtime(self):
        if not self.timearray:
            self.get_timearray()
        return self.timearray[0], self.timearray[-1]

    def get_timearray(self):
        raise NotImplementedError("Need to implement a method to read times from binary file")

    def reset_tinds(self):
        """ Reset time index and previously read data """
        self.tind = 0

    def offset(self, *args):
        raise NotImplementedError("Need to implement a method to calculate data offset in file")

    def readvar(self, *args):
        raise NotImplementedError("Need to implement a method to read binary data")

    def get_var(self, varname):
        raise NotImplementedError(
            "Need to implement a method to provide variable access by name {}".format(varname))


class TimeSeries(object):
    """ Base class for for any time series

    This is used for formatted (ascii) GENE output as well as derived data
    from GENE diagnostics
    :param objectname: Either a file name (e.g. profile_Ions.dat) or a (freely choosable) name for
     a derived dataset
    :param common: CommonData object of the current run
    """

    def __init__(self, objectname, common):
        self.objectname = objectname
        self.cm = common
        self.starttime = common.starttime
        self.endtime = common.endtime
        self.fileobject = None  # A binary file object providing data to this one
        self.timearray = []
        self.dataarray = []
        self.timeaverage = None

    def generate_timeseries(self):
        raise NotImplementedError("Need to implement a method to aggregate data")

    def generate_timeaverage(self):
        if len(self.dataarray):
            self.timeaverage = mytrapz(self.dataarray, self.timearray)
        else:
            self.generate_timeseries()
            self.timeaverage = mytrapz(self.dataarray, self.timearray)

    def get_minmaxtime(self):
        """ Return first and last time in the object

        If self.fileobject is set, try to get it from there
        """
        if np.array(self.timearray).size == 0:
            try:
                self.timearray = self.fileobject.timearray
            except AttributeError:
                raise RuntimeError("Time array not available or empty")
        return self.timearray[0], self.timearray[-1]

    def check_times(self):
        """ Set the boundaries of the time window to read"""
        first_time, last_time = self.get_minmaxtime()
       # if len(self.timearray) != len(set(self.timearray)):
        #    raise RuntimeError("Error: {} contains 2 blocks with "
         #                      "identical timestamp".format(self.objectname))
        if self.starttime == -1 or (0 < self.starttime < first_time):
            print("Using first time present in {} for starttime".format(self.objectname))
            self.starttime = first_time
        if self.endtime == -1:
            print("Using first time present in {} for endtime".format(self.objectname))
            self.endtime = first_time
        if self.starttime == -2:
            print("Using last time present in {} for starttime".format(self.objectname))
            self.starttime = last_time
        if (self.endtime == -2) or (self.endtime > last_time):
            print("Using last time present in {} for endtime".format(self.objectname))
            self.endtime = last_time
        if (self.endtime < first_time) or (self.starttime > last_time):
            raise RuntimeError("Time window not contained in {}".format(self.objectname))
        print(("starttime={}, endtime={},"
               " first_time={}, last_time={}".format(self.starttime, self.endtime, first_time,
                                                     last_time)))

    def calc_positions(self):
        """ For a single time, find element closest to given input time"""
        timearray = self.timearray
        if self.starttime == self.endtime:
            pos = np.array([bisect_left(timearray, self.starttime)])
        else:
            startpos = bisect_left(timearray, self.starttime)
            endpos = bisect_right(timearray, self.endtime)
            pos = np.arange(startpos, endpos)
        return pos


def gluetimetraces(tserieslist):
    """ Function to concatenate a list of time series (continuation runs) into one single object

    for continuation runs
    :param tserieslist: List of TimeSeries objects to connect
    :returns: One connected TimeSeries object
    """
    # TODO: Consistency checks if runs match
    if len(tserieslist) == 1:
        return tserieslist[0]
    # Use first TimeSeries as basis for the construction of the glued object
    result = tserieslist.pop(0)
    # If the timefield has been converted to a np array before we need to undo it
    try:
        result.timearray = result.timearray.tolist()
    except AttributeError:
        pass
    for tseries in tserieslist:
        # When times overlap, give following TimeSeries preference
        while result.timearray[-1] > tseries.timearray[0]:
            del result.timearray[-1]
            del result.dataarray[-1]
        result.timearray = np.concatenate((result.timearray, tseries.timearray))
        result.dataarray = np.concatenate((result.dataarray, tseries.dataarray))
    result.endtime = tserieslist[-1].endtime
    del tserieslist
    return result



"""Module containing the treatment of GENE's geometry output"""
import numpy as np
import h5py


class Geometry:
    """ Class to handle geometry input from GENE runs

    """

    def __init__(self, run):
        self.geomtype = run.pars['magn_geometry']
        if run.is_h5:
            self.getgeom_h5(run)
        elif run.is_adios:
            raise NotImplementedError("ADIOS not supported")
        else:
            self._getgeom_std(run)

    def _getgeom_std(self, run):
        #pylint: disable=invalid-name
        self.cm = run
        geom = self.getgeom(run)
        self.gxx = geom[0]
        self.gxy = geom[1]
        self.gxz = geom[2]
        self.gyy = geom[3]
        self.gyz = geom[4]
        self.gzz = geom[5]
        self.Bfield = geom[6]
        self.dBdx = geom[7]
        self.dBdy = geom[8]
        self.dBdz = geom[9]
        self.jacobian = geom[10]
        self.R = geom[11]
        self.Z = geom[12]
        self.dxdR = geom[13]
        self.dxdZ = geom[14]

        if not run.x_local:
            self.Cy = geom[15, 0]
            self.Cxy = geom[16, 0]
            self.q = geom[17, 0]
            self.dpdx_pm_arr = geom[18, 0]
            self.jaco3d = np.broadcast_to(self.jacobian[:, np.newaxis, :],
                                          (run.pnt.nz0, run.pnt.nky0, run.pnt.nx0))
        elif not run.y_local:
            self.jaco3d = np.broadcast_to(self.jacobian[:, :, np.newaxis],
                                          (run.pnt.nz0, run.pnt.nky0, run.pnt.nx0))

    @staticmethod
    def untangle_1d(arrlist, start, nxorny):
        """ Process a geometry file section that contains a 1d (x or y) field

        :param arrlist: The entire file as a list of lines
        :param start: The starting index of the quantity in arrlist
        :param nxorny: nx0 or nky0 depending on the type of simulation
        :returns: A 1d numpy array
        """
        arr_1d = np.zeros(nxorny)
        i = 0
        for line in arrlist[start:-1]:
            for j in range(len(line.split())):
                arr_1d[i + j] = line.split()[j]
            i += len(line.split())
            if i >= nxorny:
                break
        return arr_1d

    @staticmethod
    def untangle_2d(arrlist, start, nxorny, nz):
        """ Process a geometry file section that contains a 2d (x or y, z) field

        :param arrlist: The entire file as a list of lines
        :param start: The starting index of the quantity in arrlist
        :param nxorny: nx0 or nky0 depending on the type of simulation
        :param nz: nz0, number of z grid points
        :returns: A 2d numpy array
        """
        arr_2d = np.zeros((nz, nxorny))
        ik = 0
        for line in arrlist[start:-1]:
            arr = line.split()[:]
            lb = ik%nxorny
            ub = min(lb + 16, nxorny)
            k = int(ik/nxorny)
            dim = ub - lb
            arr_2d[k, lb:ub] = arr[0:dim]
            if dim < 16:
                arr_2d[k + 1, 0:16 - dim] = arr[dim:]
            ik += 16
            if ik >= nz*nxorny:
                break
        return arr_2d

    def getgeom(self, run):
        """ Returns the geometry from a non-hdf5 file """
        local = run.x_local and run.y_local
        with open(run.in_folder+"/{}{}".format(self.geomtype.strip("'"),
                                                run.fileextension), "r") as geomfile:
            if local:
                geom = np.empty((16, run.pnt.nz0), dtype=np.float64)
                k = 0
                for line in geomfile:
                    if len(line.split()) == 16:
                        geom[:, k] = line.split()[:]
                        k += 1
                    elif line.startswith('Cy'):
                        self.Cy = float(line.split()[-1])
                    elif line.startswith('Cxy'):
                        self.Cxy = float(line.split()[-1])
            elif run.y_local:  # x-global
                geom = self.getgeom_glob(geomfile, run.pnt.nz0, run.pnt.nx0)
            elif run.x_local:  # y-global
                geom = self.getgeom_glob(geomfile, run.pnt.nz0, run.pnt.nky0)
            else:
                raise NotImplementedError("xy global not supported")
        return geom

    def getgeom_glob(self, geomfile, nz0, nxorny):
        """ Subroutine for geometry files from global runs """
        NUMFIELDS = 19
        geom = np.zeros((NUMFIELDS, nz0, nxorny), dtype=np.float64)

        geomdict = {'gxx': 0, 'gxy': 1, 'gxz': 2, 'gyy': 3, 'gyz': 4, 'gzz': 5, 'Bfield': 6,
                    'dBdx': 7, 'dBdy': 8, 'dBdz': 9, 'jacobian': 10, 'geo_R': 11, 'geo_Z': 12,
                    'geo_c1': 13, 'geo_c2': 14, 'C_y': 15, 'C_xy': 16, 'q': 17, 'dpdx_pm_arr': 18}
        pos_start = np.zeros(NUMFIELDS, dtype=int)  # where a field starts
        num = 0
        parheader = True
        geomlist = geomfile.readlines()
        for linnum, line in enumerate(geomlist):
            if parheader:  # ignore the top part of the file (copy of the geometry namelist)
                if line.startswith(r'/'):
                    parheader = False
                continue
            if len(line.split()) == 1:  # the variable names
                try:
                    num = geomdict[line.strip()]
                    pos_start[num] = linnum + 1
                except KeyError:
                    try:  # Test if it is a single number (can occur for 1d arrays)
                        float(line.strip())
                    except ValueError:
                        raise RuntimeError("Unknown entry name in geometry file")
        for num in range(NUMFIELDS):
            if pos_start[num] == 0:  # This should only occur if a field does not exist
                continue
            if num in [15, 16, 17, 18]:  # We have a 1d field following
                geom[num] = self.untangle_1d(geomlist, pos_start[num], nxorny)
            else:
                geom[num] = self.untangle_2d(geomlist, pos_start[num], nxorny, nz0)

        return geom

    def getgeom_h5(self, run):
        """ Returns the geometry from a hdf5 file """

        geomfile = run.in_folder + "/{}{}".format(self.geomtype.strip("'"), run.fileextension)
        self.cm = run
        geom = h5py.File(geomfile, 'r')

        self.gxx = geom.get('/metric/g^xx').value
        self.gxy = geom.get('/metric/g^xy').value
        self.gxz = geom.get('/metric/g^xz').value
        self.gyy = geom.get('/metric/g^yy').value
        self.gyz = geom.get('/metric/g^yz').value
        self.gzz = geom.get('/metric/g^zz').value
        self.Cy = geom.get('/metric/C_y').value
        self.Cxy = geom.get('/metric/C_xy').value

        self.Bfield = geom.get('/Bfield_terms/Bfield').value
        self.dBdx = geom.get('/Bfield_terms/dBdx').value
        self.dBdy = geom.get('/Bfield_terms/dBdy').value
        self.dBdz = geom.get('/Bfield_terms/dBdz').value
        self.jacobian = geom.get('/Bfield_terms/Jacobian').value

        if not (not run.x_local and not run.y_local):
            self.R = geom.get('/shape/R').value
            self.Z = geom.get('/shape/Z').value
            self.dxdR = geom.get('/shape/dxdR').value
            self.dxdZ = geom.get('/shape/dxdZ').value

        try:
            self.x_cart = geom.get('/cartesian_cords/x_cart').value
            self.y_cart = geom.get('/cartesian_cords/y_cart').value
            self.z_cart = geom.get('/cartesian_cords/z_cart').value
        except:
            pass

        try:
            self.q = geom.get('/profile/q_prof').value
        except AttributeError:
            self.q = run.pnt.q0

        try:
            self.dpdx_arr = geom.get('/profile/dpdx_pm_arr').value
        except AttributeError:
            self.dpdx_arr = run.pnt.dpdx_pm

        geom.close()



""" Base module for plotting

Contains:
- Matplotib imports and settings

"""

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.font_manager import FontProperties
import matplotlib as mpl
from pkg_resources import parse_version

if parse_version(mpl.__version__) >= parse_version("1.5"):
    from cycler import cycler


class Plotting(object):
    """ Base class for plotting routines

    Contains the parameters used for the plots such as
    line colors, and font sizes
    Sets the line and point marker widths/sizes
    Sets the default color cycle (respecting an API change in matplotlib 1.5)
    """

    def __init__(self):
        # Matplotlib 2.x changes the default plot styles
        if parse_version(mpl.__version__) >= parse_version("2.0"):
            # Colormap for quantities >0 (e.g. heat flux)
            self.cmap_unidirect = mpl.cm.inferno
            mpl.rc('font', size=15)
        else:
            self.cmap_unidirect = mpl.cm.hot
            mpl.rc('font', size=20)
        # for quantities with critical value (e.g. 0 for phi)
        self.cmap_bidirect = mpl.cm.bwr
        self.color_list = plt.cm.Dark2(np.linspace(0, 1.0, 9))  # Plot line colors
        # Set some basic plot properties
        # Those are set globally everytime a Plotting object is instantiated!
        mpl.rc('legend', fontsize="small")
        mpl.rc('lines', linewidth=2, markeredgewidth=1.5, markersize=10)
        # This sets the figure frames to transparent both on screen and in calls to savefig
        mpl.rc('figure', facecolor=(1, 1, 1, 0))
        mpl.rc('figure', frameon=False)
        mpl.rc('savefig', facecolor=(1, 1, 1, 0))
        mpl.rc('savefig', frameon=False)
        # mpl 1.5 introduces a new, more flexible prop_cycle parameter, so different line styles
        # can be defined in the else case as well
        if parse_version(mpl.__version__) <= parse_version("1.4"):
            mpl.rc('axes', color_cycle=self.color_list)
        else:
            mpl.rc('axes', prop_cycle=cycler('color', self.color_list))
        # Dictionary that converts internal names to well-formatted LaTeX for the plots
        self.titles = {'omts': r'$a/L_T$', 'omns': r'$a/L_n$', 'ns': r'$n$', 'Ts': r'$T$',
                       'Gammaturb': r'$\Gamma_{\mathrm{turb}}/\Gamma_{gB}$',
                       'Gammanc': r'$\Gamma_{\mathrm{NC}}/\Gamma_{gB}$',
                       'Qturb': r'$Q_{\mathrm{turb}}/Q_{gB}$',
                       'Qnc': r'$Q_{\mathrm{NC}}/Q_{gB}$',
                       'Piturb': r'$\Pi_{\mathrm{turb}}/\Pi_{gB}$',
                       'Pinc': r'$\Pi_{\mathrm{turb}}/\Pi_{gB}$',
                       "jbs": r'$j_\mathrm{BS}$', "phi": r"$\phi$",
                       "apar": r"$A_\parallel$", "dens": r"$n$",
                       "tpar": r"$T_\parallel$", "tperp": r"$T_\perp$",
                       "qpar": r"$q_\parallel + 1.5p_0 u_\parallel$",
                       "qperp": r"$q_\perp + p_0 u_\parallel$", "upar": r"$u_\parallel$",
                       "bpar": r"$B_\parallel$", "n00": r"$N_{00}$",
                       "n20": r"$N_{20}$", "n02": r"$N_{02}$"}

    @staticmethod
    def _minlogaxis(arr):
        """ Calculate lower limit for plotting on a log axis

        lower Integer * 10^minimumexponent

        :param arr: Array to base the limit on
        """
        minim = arr[arr != 0].min()  # ignore 0
        min_exp = np.floor(np.log10(minim))
        min_mant = np.floor(minim*10**(-min_exp))
        floormin = min_mant * 10**min_exp
        return floormin

    @staticmethod
    def _maxlogaxis(arr):
        """ Calculate upper limit for plotting on a log axis

         higher Integer * 10^maximumexponent

        :param arr: Array to base the limit on
        """
        maxim = arr[arr != 0].max()
        max_exp = np.floor(np.log10(maxim))
        max_mant = np.ceil(maxim*10**(-max_exp))
        ceilmax = max_mant * 10**max_exp
        return ceilmax

    @staticmethod
    def gyrobohm_SI(common, quantity):
        """ Convert gyroBohm unit to SI unit

        :param common: CommonData object
        :param quantity: Name of the quantity to calculate gB for
        """
        pnt = common.pnt
        elementary_charge = 1.60217662e-19
        temp_unit = 1e3*elementary_charge  # temp is in keV
        dens_unit = 1e19  # m**(-3)
        mass_unit = 1.6726219e-27  # proton mass
        # Magnetic field and reference length are already in T and m
        nref = pnt.nref*dens_unit
        Tref = pnt.Tref*temp_unit
        mref = pnt.mref*mass_unit
        cref = np.sqrt(Tref/mref)
        rhoref = mref*cref/elementary_charge/pnt.Bref
        Gamma_gB = nref*cref*rhoref ** 2/pnt.Lref**2
        gBunits = {"Ts": Tref, "ns": nref, "omts": 1, "omns": 1, "Gammanc": Gamma_gB/1e19,
                   "Gammaturb": Gamma_gB/1e19, "Qturb": Gamma_gB*Tref/1e3, "Qnc": Gamma_gB*Tref/1e3,
                   "Pinc": Gamma_gB*cref*mref, "Piturb": Gamma_gB*cref*mref,
                   "jbs": pnt.Bref*nref*cref*rhoref/pnt.Lref}
        return gBunits[quantity]

    @staticmethod
    def round_to_n(x, n):
        # Helper function to round to n significant digits
        return np.round(x, -int(np.floor(np.log10(np.abs(x)))) + (n - 1))


""" errors.py: Collection of routines to calculate error estimates and other statistical tools

"""
import numpy as np
from multiprocessing.dummy import Pool as ThreadPool
from .averages import mytrapz


def autocorrtime(data, timefld):
    """ Calculate autocorrelation time of data array

    :param data: First dimension is considered the direction to correlate over
    :param timefld: The domain to correlate in
    :returns: 1/e time of the autocorrelation function in an array
    that has ndim(data)-1 dimensions
    """
    data = np.array(data, dtype=float)
    if len(timefld) <= 2:
        return np.zeros(data.shape[1:])
    if data.ndim == 1:
        return autocorrtime_1d(data, timefld)
    else:
        return autocorrtime_nd(data, timefld)


def autocorrtime_1d(data, timefld):
    """ Correlation time calculation for 1d arrays (only time)

    :param data: 1d array of the data to correlate
    :param timefld: 1d array of the correlation domain
    :returns: 1/e correlation time
    """
    meanvar = mytrapz(data, timefld)
    # The mean needs to be substracted to arrive at
    # the statistical definition of the autocorrelation
    data -= meanvar
    result = np.correlate(data, data, mode="full")
    result = result[int(result.size/2):]
    overe = np.squeeze(np.where(result > np.exp(-1)*result[0]))
    if overe.size <= 1:
        corrtime = 0
    else:
        cort_ind = np.array_split(overe, np.where(np.diff(overe) != 1)[0] + 1)[0][-1]
        corrtime = timefld[cort_ind] - timefld[0]
    return corrtime


def autocorrtime_nd(data, timefld):
    """ Correlation time calculation for >1d arrays

    :param data: array of the data to correlate
    :param timefld: array of the correlation domain
    :returns: 1/e correlation times in a numpy array
    """
    flat = data.reshape(data.shape[0], -1)
    pool = ThreadPool()
    corrtimes = np.array(pool.map(lambda ind: _autocorrtime_nd_helper(ind, flat, timefld, data),
                                  np.ndindex(data.shape[1:])))
    # This parallelises the following loop
    # corrtimes = np.zeros(data.shape[1:])
    # for ind in np.ndindex(data.shape[1:]):
    #     corrtimes[ind] = _autocorrtime_nd_helper(ind, flat, timefld, data)
    pool.close()
    pool.join()
    return corrtimes.reshape(data.shape[1:])


def _autocorrtime_nd_helper(ind, flat, timefld, data):
    var = np.squeeze(flat[:, np.ravel_multi_index(ind, data.shape[1:])])
    return autocorrtime_1d(var, timefld)


def windowerr(data, timefld, std_mean=True, n_win=0):
    """ Wrapper for 1d or multidimensional windowed error calculation

    :param data: array of the data to calculate the uncertainty
    :param timefld: array of the sampling times for data
    :param std_mean: switches to error of the mean with student-t correction for
    low n_win instead of standard deviation of the windowed sample
    :param n_win: Number of windows to group the data. 0 means automatic calculation based on the
    autocorrelation time (recommended default)
    :returns: The windowed error and the autocorrelation time.
    """
    data = np.array(data)
    if n_win <= 0:
        ctimes = autocorrtime(data, timefld)
    else:  # Assign a dummy array for the correlation times
        ctimes = np.zeros(data.shape[1:])
    if data.ndim == 1:
        return windowerr_1d(data, timefld, ctimes, std_mean, n_win), ctimes
    else:
        return windowerr_nd(data, timefld, ctimes, std_mean, n_win), ctimes


def windowerr_1d(data, timefld, corrt, std_mean=True, n_win=0):
    """ Standard deviation of the windowed mean values

    Tries to use window widths of 5 corrt, fallback are at least 2
    std_mean switches to error of the mean
    with student-t correction for low n_win

    """
    if len(timefld) <= 2:
        # print("Not enough times for error calculation")
        return 0
    total_dt = timefld[-1] - timefld[0]
    # If no window number is given, use correlation time input
    # start with window width = 5 tcorr; if fewer than 6, reduce width, min 2
    # same as in IDL diag, tools.pro, stat_avg_err,
    # except the minimal n_win is 10 there
    if n_win <= 0:
        if total_dt >= 30*corrt:
            if np.allclose(corrt, 0):
                return 0
            else:
                n_win = int(np.floor(total_dt/(5*corrt)))
        else:
            if total_dt >= 12*corrt:
                n_win = 6
            else:
                # print('Not enough windows')
                return 0
    win_inds = []
    wwidth = total_dt/n_win
    for iwin in range(n_win):
        win_inds.append(np.squeeze(np.where(
            (timefld - timefld[0] >= iwin*wwidth)&(timefld - timefld[0] < (iwin + 1)*wwidth))))
    means = np.empty(n_win)
    for iwin, window in enumerate(win_inds):
        if np.array(window).size < 1:
            print('windowerr_1d: windows with < 1 steps, aborting error computation')
            return 0
        means[iwin] = mytrapz(data[window], timefld[window])
    std_error = np.std(means, ddof=1)
    if std_mean:
        std_error *= np.sqrt(1./(n_win - 2))
    return std_error


def windowerr_nd(data, timefld, ctimes, std_mean=True, n_win=0):
    """ Standard error of the windowed mean value

    Function for multidimensional data which calls the 1d version repeatedly
    """
    std = np.zeros(data.shape[1:])
    flat = data.reshape(data.shape[0], -1)
    for ind in np.ndindex(data.shape[1:]):
        var = np.squeeze(flat[:, np.ravel_multi_index(ind, data.shape[1:])])
        std[ind] = windowerr_1d(var, timefld, ctimes[ind], std_mean, n_win)
    return std



""" Contains the class to read binary srcmom files"""

from os.path import getsize
import numpy as np

from .base_file import BinaryFile, TimeSeries


class SrcmomFile(BinaryFile):
    """ Class to read a binary srcmom output file

    A srcmom file entry consists of the time stamp followed by radial profiles for (1)particle,
     (2)momentum and (3)heat input by (1)the Krook heat source, (2)the Krook particle source and (3)
     the f0 contributions (both the localized source and the neoclassical term)

    :param filename: Typically srcmom_spec_fext
    :param common: CommonData object of the run
    """

    def __init__(self, filename, common, usemmap=False):
        super().__init__(filename, common, usemmap=usemmap)
        self.prev_data = (0, "", "")
        self.filename = filename
        self.srcdata = None
        self._set_sizes()

    def _set_sizes(self):
        """ Set up the sizes for the records in Fortran binary files"""
        super()._set_sizes()
        self.nmoms = 3
        self.nsrc = 3
        self.moms = {"part": 0, "mom": 1, "energy": 2}
        self.sources = {"ck_heat": 0, "ck_part": 1, "f0": 2}
        # Each dataset at a time consists of 3 records (the source types) which contain 3 nx arrays
        # (particle, heat, momentum moment). Each record is surrounded by its length as a 4 byte int
        self.entrysize = self.cm.pnt.nx0*self.realsize
        self.leapsrcmom = self.nmoms*(self.nsrc*self.entrysize + 2*self.intsize)

    def get_timearray(self):
        """ Get the time array from the file"""
        # Fortran has a particular binary format: see time_entry()
        for i in range(int(getsize(self.filename)/(self.leapsrcmom + self.tesize))):
            self.timearray.append(float(self.te.unpack(self.file.read(self.tesize))[1]))
            self.file.seek(self.leapsrcmom, 1)

    def reset_tinds(self):
        """ Reset time index and previously read data """
        super().reset_tinds()
        self.prev_data = (0, "", "")

    def readvar(self, var, subvar):
        """ Read one of the nine vectors at a given time

        :param var: one of self.moms values
        :param subvar: one of self.sources values
        """
        if not (self.tind, var, subvar) == self.prev_data:
            self.srcdata = self.readfunc(self.file, count=self.cm.pnt.nx0, dtype=self.nprt,
                                         offset=self.offset(var, subvar))
            self.prev_data = (self.tind, var, subvar)
        return self.srcdata

    def offset(self, var, subvar):
        """ Calculate offset in field file for a given timestep and variable"""
        return self.tind*(self.tesize + self.leapsrcmom) + self.tesize + self.intsize + subvar*(
            self.nmoms*self.entrysize + 2*self.intsize) + var*self.entrysize

    def srcmoment(self, momtype, srctype):
        """ Return the source moment, read from the file if necessary

        :param momtype: Type of the moment (key of self.moms)
        :param srctype: Type of the source (key of self.source)
        """
        try:
            imom = self.moms[momtype]
            isrc = self.sources[srctype]
        except KeyError:
            raise KeyError("Not a valid source or source moment entry")
        return self.readvar(imom, isrc)

    def get_var(self, varname):
        mom, src = varname.split("_")
        return self.srcmoment(mom, src)

class SrcmomSeries(TimeSeries):
    """ Construct a time series of the source moment diag """

    def __init__(self, common, rundatafiles, momtype, srctype):
        super().__init__("srcmomseries", common)
        self.momtype = momtype
        self.srctype = srctype
        self.srcmoms = [[] for _ in common.specnames]
        self.rundatafiles = rundatafiles

    def generate_timeseries(self):
        srcmomfiles = [self.rundatafiles.get_fileobject("srcmom_{}".format(spec))
                       for spec in self.cm.specnames]
        self.fileobject = srcmomfiles[0]
        self.check_times()
        tinds = self.calc_positions()
        for ispec, smf in enumerate(srcmomfiles):
            for pos in tinds:
                smf.set_tind(pos)
                self.srcmoms[ispec].append(np.array(smf.srcmoment(self.momtype, self.srctype)))
        self.timearray = np.array(self.timearray)[tinds]



""" profiledata.py: handles background profile information"""
import numpy as np
import h5py
import os
import matplotlib.pyplot as plt
import matplotlib.cm as cm

class ProfileData:
    """Class handling the profile files of a run
    """
    def __init__(self,folder,extension,params):
        """ Read parameter file and create empty arrays for profile data """
        self.species=params.species
        self.nref=params.pnt.nref
        self.Tref=params.pnt.Tref
        self.xs = np.empty(params.pnt.nx0,dtype=float)
        self.x_a = np.empty(params.pnt.nx0,dtype=float)
        self.temp = np.empty(params.pnt.n_spec,dtype=float)
        self.dens = np.empty(params.pnt.n_spec,dtype=float)
        self.T0 = np.empty([params.pnt.n_spec, params.pnt.nx0],dtype=float)
        self.n0 = np.empty([params.pnt.n_spec, params.pnt.nx0],dtype=float)
        self.omt0 = np.empty([params.pnt.n_spec, params.pnt.nx0],dtype=float)
        self.omn0 = np.empty([params.pnt.n_spec, params.pnt.nx0],dtype=float)
        self.__set_background_profiles(folder,extension,params)
        
    def __set_background_profiles(self,folder,extension,params):
        for i_spec in range(params.pnt.n_spec):
            if params.pnt.x_local:
                self.T0[:,i_spec]=1.0
                self.n0[:,i_spec]=1.0
                self.omt0[:,i_spec]=params.species[i_spec]['omt']
                self.omn0[:,i_spec]=params.species[i_spec]['omn']
            else:
                self.__read_input_profile(folder,extension,i_spec, params)

    def __read_input_profile(self,folder,extension,i_spec, params):
        """Get the initial profiles from profiles_spec_fileextension"""
        if  params.pnt.write_h5:
            for n, spec in enumerate(self.specnames):
                proffile = self.folder / 'profiles_{}{}'.format(spec, extension)
                prof = h5py.File(proffile, 'r')
                self.x_a = prof.get("/position/x_o_a")[()]
                self.xs = prof.get("/position/x_o_rho_ref")[()]
                self.T0s[:, n] = prof.get('/temp/T')[()]
                self.omt0s[:, n] = prof.get('/temp/omt')[()]
                self.n0s[:, n] = prof.get('/density/n')[()]
                self.omn0s[:, n] = prof.get('/density/omn')[()]
                prof.close()
        else:
            self.dens[i_spec]=params.species[i_spec]['dens']
            self.temp[i_spec]=params.species[i_spec]['temp']
            file=os.path.join(folder,'profiles_{}{}'.format(params.species[i_spec]['name'],extension))
            with open(file,'r') as pfile:
                lines = pfile.readlines()
                for i in range(2, params.pnt.nx0+2):
                    self.x_a[i-2] = float(lines[i].split()[0])
                    self.xs[i-2] = float(lines[i].split()[1])
                    self.T0[i_spec,i-2] = float(lines[i].split()[2])/params.species[i_spec]['temp']/params.pnt.Tref
                    self.n0[i_spec,i-2] = float(lines[i].split()[3])/params.species[i_spec]['dens']/params.pnt.nref
                    self.omt0[i_spec,i-2] = float(lines[i].split()[4])
                    self.omn0[i_spec,i-2] = float(lines[i].split()[5])

    def show(self, use_SI=False, spec=None, specname=None):
        def specplot(i_spec):
            fig = plt.figure(figsize=(12,4), dpi= 100, facecolor='w', edgecolor='k')
            ax = fig.add_subplot(1,4,1)
            ax.plot(self.x_a, self.T0[i_spec,:]*self.species[i_spec]['temp']*(self.Tref if use_SI else 1),
                    '-b',label=(r"T [$keV$]" if use_SI else r'$T/T_{\rm ref}$'))
            ax.legend();ax.set_title(r"${}$".format(self.species[i_spec]['name']));ax.set_xlabel(r"$x/a$");
            ax = fig.add_subplot(1,4,2)
            ax.plot(self.x_a, self.n0[i_spec,:]*self.species[i_spec]['dens']*(self.nref if use_SI else 1),
                    '-b',label=(r"n [$10^{19}/m^3$]" if use_SI else r'$n/n_{\rm ref}$'))
            ax.legend();ax.set_title(r"${}$".format(self.species[i_spec]['name']));ax.set_xlabel(r"$x/a$");            
            ax.autoscale(tight=True);
            ax = fig.add_subplot(1,4,3)
            ax.plot(self.x_a, self.omt0[i_spec,:],'-b',label=r'$\omega_{T}$')
            ax.legend();ax.set_title(r"${}$".format(self.species[i_spec]['name']));ax.set_xlabel(r"$x/a$");
            ax.autoscale(tight=True);
            ax = fig.add_subplot(1,4,4)
            ax.plot(self.x_a, self.omn0[i_spec,:],'-b',label=r'$\omega_{n}$')
            ax.legend();ax.set_title(r"${}$".format(self.species[i_spec]['name']));ax.set_xlabel(r"$x/a$");
            ax.autoscale(tight=True);            
            fig.tight_layout()

        if spec and spec<len(self.species):
            specplot(spec)
        elif specname:
            for i_spec in range(self.T0.shape[0]):
                if self.species[i_s]['name']==specname:
                    specplot(i_spec)
        else:    
            for i_spec in range(self.T0.shape[0]):
                specplot(i_spec)


# -*- coding: utf-8 -*-
"""
Main script to handle uploading a set of runs with same metadata to the MGK database
                    
@author: Austin Blackmon, Dongyang Kuang, Venkitesh Ayyar
"""

import sys
import os
import argparse
from sys import exit
from bson.objectid import ObjectId

from mgkdb.support.mgk_file_handling import get_suffixes, upload_to_mongo, isLinear, Global_vars, f_get_linked_oid, f_set_metadata, f_load_config, f_user_input_metadata
from mgkdb.support.mgk_login import mgk_login,f_login_dbase

import yaml 

def f_parse_args():
    #==========================================================
    # argument parser
    #==========================================================
    parser = argparse.ArgumentParser(description='Process input for uploading files')

    parser.add_argument('-T', '--target', help='Target run output folder')
    parser.add_argument('-SIM', '--sim_type', choices=['GENE','CGYRO','TGLF','GS2','GX'], type=str, help='Type of simulation', required=True)
    parser.add_argument('-A', '--authenticate', default = None, help='locally saved login info, a .pkl file')
    parser.add_argument('-C', '--config_file', default = None, help='Configuration file (.yaml) to avoid terminal prompts.')
    parser.add_argument('-D', '--default', default = False, action='store_true', help='Using default inputs for all.')
    parser.add_argument('-V', '--verbose', dest='verbose', default = False, action='store_true', help='output verbose')
    
    parser.add_argument('-Ex', '--extra', dest='extra', default = False, action='store_true', help='whether or not to include any extra files for each suffix')
    parser.add_argument('-L', '--large_files', dest='large_files', default = False, action='store_true', help='whether or not to include large files')
    
    return parser.parse_args()

def main_upload(target, default, sim_type, extra, authenticate, verbose, large_files, config_file):
    '''
    Upload a set of suffixes with common Metadata
    '''
    ### Initial setup 
    upload_folder = os.path.abspath(target)
    global_vars = Global_vars(sim_type)    

    ### Connect to database 
    login = f_login_dbase(authenticate)
    client, database = login.connect()
    with client:
        user = login.login['user']

        # manual_time_flag = not default
        manual_time_flag = False

        print(f'Scanning in {upload_folder} *******************\n')
        linear = isLinear(upload_folder, sim_type)                      
        all_suffixes = get_suffixes(upload_folder, sim_type)

        if not all_suffixes:
            print("Did not find any suffixes in the folder",upload_folder)
            return

        if not default:
            if config_file is not None:
                config_dict = f_load_config(config_file)

                ## to do : Add check for format of config file

                user_input = config_dict['user_input']    
                metadata_info = config_dict['metadata']
                shared_files = user_input['shared_files']

                if user_input['extra_files']:
                    ex_files = user_input['extra_files'].split(',')
                    global_vars.Docs_ex +=ex_files

                ## Add suffixes
                ip_suffixes = user_input['suffixes']
                if ip_suffixes: 
                    suffixes=ip_suffixes.split(',')
                    ## Check if input suffixes exist in the folder
                    incorrect_suffixes = [s1 for s1 in suffixes if s1 not in all_suffixes]
                    if incorrect_suffixes: 
                        print("The following suffixes provided don't exist in this folder")
                        print(f"Found in {upload_folder} these suffixes:\n {all_suffixes}")
                        print('Skipping this folder')
                else:    suffixes = None

                ## Adding shared files info
                run_shared = shared_files.split(',') if shared_files else None
                
                ## Add metadata info
                metadata = metadata_info
                metadata['CodeTag']['sim_type']=sim_type
                linked_id_strg = metadata['DBtag']['linkedObjectID']
                if linked_id_strg is not None:
                    metadata['DBtag']['linkedObjectID'] = f_get_linked_oid(database, linked_id_strg)

                no_prompts         = user_input['no_prompts']
                reupload_if_exists = user_input['reupload_if_exists']

            else: ## Get data through user input 
                if extra: # this will change the global variable
                    ex_files = input('Please type FULL file names to update, separated by comma.\n').split(',')
                    global_vars.Docs_ex +=ex_files

                print("Found in {} these suffixes:\n {}".format(upload_folder, all_suffixes))
                
                suffixes = input('Which run do you want to upload? Separate them by comma. \n Press q to skip. Press ENTER to upload ALL.\n')
                if suffixes == 'q':
                    print("Skipping the folder {}.".format(upload_folder))
                elif len(suffixes):
                    suffixes = suffixes.split(',')
                else:
                    suffixes = None                              
                
                run_shared = input('Do you want to upload any shared files for all suffixes? Please specify path relative to parent folder.\n Separate them by comma. Press Enter to skip.\n')
                if len(run_shared):
                    run_shared = run_shared.split(',')
                else:
                    run_shared = None

                ### Metadata inputs
                user_ip_dict = f_user_input_metadata(database)
                metadata = f_set_metadata(**user_ip_dict,user=user, sim_type=sim_type)

                no_prompts = False
                reupload_if_exists = False
        else:
            suffixes = None
            run_shared = None
            metadata = f_set_metadata(user=user, sim_type=sim_type)
            no_prompts = True
            reupload_if_exists = False
            
        upload_to_mongo(database, linear, metadata, upload_folder, suffixes, run_shared,
                        large_files, verbose, manual_time_flag, global_vars, no_prompts=no_prompts, reupload_if_exists=reupload_if_exists)

def main():

    ### Parse arguments 
    args = f_parse_args()
    input_args = vars(args)
    print(input_args)

    main_upload(**input_args)

## Runner 
if __name__=="__main__":
    main()

# Example command : 
## python mgk_uploader.py -A <fname.pkl> -T test_data/test_gene1_tracer_efit -SIM GENE -C template_user_input.yaml



""" Module for plotting of x-y contours"""

import matplotlib.pyplot as plt
import matplotlib.animation as animation
import numpy as np
import warnings
import multiprocessing

import .baseplot
from ..data.slices import MomFieldSlice
from ..utils.comm import DiagSpace

import ..utils.averages as averages
import ..utils.geom


class MomFieldContour(object):
    """ Create a representation of a x-y map for a mom or field quantity"""

    def __init__(self, common, species, rundatafiles, zind=None, zavg=False, quantities=("phi",)):
        if quantities is None:
            raise RuntimeError("No quantities given for contours")
        if not zind:
            zind = int(common.pnt.nz0/2)
        zrange = (zind, zind + 1)
        diagspace = DiagSpace(common.spatialgrid, False, False, False, (None, None), (None, None),
                              zrange, False, False, zavg)
        self.species = species
        self.cm = common
        self.contourdata = {}
        self.fetch_data(diagspace, quantities, rundatafiles)

    def fetch_data(self, diagspace, quantities, rundatafiles):
        if set(quantities).intersection({"phi", "apar", "bpar"}) and set(quantities).intersection(
                {"dens", "tpar", "tperp", "qpar", "qperp", "upar"}):
            bothmomandfield = True
        else:
            bothmomandfield = False
        for quantity in quantities:
            qslice = MomFieldSlice(self.cm, quantity, species=self.species, diagspace=diagspace,
                                   rundatafiles=rundatafiles, modifier_func=lambda dum: dum)

            if bothmomandfield and quantity in ["phi", "apar", "bpar"]:
                sparse_field = int(self.cm.pnt.istep_mom/self.cm.pnt.istep_field)
                qslice.generate_timeseries(sparsefactor=sparse_field)
            else:
                qslice.generate_timeseries()
            self.contourdata.update({quantity: qslice})


class PlotXYContour(pydiag.diagplots.baseplot.Plotting):
    """ Class to handle contour plots of mom or field data in x-y real space"""

    def __init__(self, contourlist):
        super().__init__()
        self.contourlist = contourlist
        self.anims = []  # The animation objects need to be stored for the animations to run
        self.figlist = []
        self.axlist = []
        self.axcblist = []

    def createfigs(self, animate=False, save=False):
        for cont in self.contourlist:
            for ic, contdata in enumerate(cont.contourdata):
                if not animate:
                    cont.contourdata[contdata].generate_timeaverage()
                    absval = int(
                            self.round_to_n(np.max(np.abs(cont.contourdata[contdata].timeaverage)),
                                            2))
                    fig = plt.figure()
                    ax = fig.add_subplot(111)
                    cm1 = ax.contourf(cont.cm.spatialgrid.x, cont.cm.spatialgrid.y,
                                      np.squeeze(cont.contourdata[contdata].timeaverage).T, 100,
                                      cmap=self.cmap_bidirect, vmax=absval, vmin=-absval,
                                      zorder=-20)
                    ax.set_rasterization_zorder(z=-10)
                    ax.set_xlabel(r"$x/\rho_{ref}$")
                    ax.set_ylabel(r"$y/\rho_{ref}$")
                    ax.set_title(self.titles[contdata])
                    fig.colorbar(cm1)
                    fig.tight_layout()
                    if save:
                        fig.savefig("contour_{}{}_{}.pdf".format(contdata, cont.species,
                                                                 cont.cm.fileextension))
                else:
                    absval = int(
                        self.round_to_n(np.max(np.abs(cont.contourdata[contdata].dataarray)), 2))
                    self.figlist.append(plt.figure())
                    self.axlist.append(self.figlist[-1].add_subplot(111))

                    cma = self.axlist[-1].contourf(cont.cm.spatialgrid.x, cont.cm.spatialgrid.y,
                                                   np.squeeze(cont.contourdata[contdata].dataarray[
                                                                  0]).T, 100,
                                                   cmap=self.cmap_bidirect, vmax=absval,
                                                   vmin=-absval, zorder=-20)
                    self.axcblist.append(self.figlist[-1].colorbar(cma).ax)

                    def on_click(event):
                        if event.dblclick:
                            for ani in self.anims:
                                if ani.running:
                                    ani.event_source.stop()
                                else:
                                    ani.event_source.start()
                                ani.running ^= True

                    def animate(t, axs=None, fig=None, axcb=None, contdata=None):
                        axs.clear()
                        axcb.clear()
                        axs.set_xlabel(r"$x/\rho_{ref}$")
                        axs.set_ylabel(r"$y/\rho_{ref}$")
                        axs.set_rasterization_zorder(z=-10)
                        axs.set_title(r"{}, t={:.3f}".format(self.titles[contdata],
                                                         cont.contourdata[contdata].timearray[t]))
                        cma = axs.contourf(cont.cm.spatialgrid.x, cont.cm.spatialgrid.y,
                                           np.squeeze(cont.contourdata[contdata].dataarray[t]).T,
                                           100, cmap=self.cmap_bidirect, vmax=absval, vmin=-absval,
                                           zorder=-20)
                        cb = fig.colorbar(cma, cax=axcb)
                        return cma, cb

                    self.anims.append(animation.FuncAnimation(self.figlist[-1], animate, range(
                            len(cont.contourdata[contdata].timearray)), fargs=(
                        self.axlist[ic], self.figlist[ic], self.axcblist[ic], contdata),
                                                              blit=False))
                    for anim in self.anims:
                        anim.running = True
                    for fig in self.figlist:
                        fig.canvas.mpl_connect('button_press_event', on_click)
                    if save:
                        # Keep the plotting from freezing by doing the saving in a subprocess
                        def save_anim():
                            print("Saving animation in the background, this can take a long while")
                            self.anims[-1].save(
                                    "animatedcontour_{}_{}{}.mp4".format(contdata, cont.species,
                                                                         cont.cm.fileextension),
                                    dpi=300)
                            print("Finished saving animation")
                        p = multiprocessing.Process(target=save_anim)
                        p.daemon = True
                        p.start()



# -*- coding: utf-8 -*-

from tkinter import END
import matplotlib.pyplot as plt
import numpy as np
from ..putils.averages import mytrapz, xy_av3d
from .baseplot import Plotting
from .diagnostic import Diagnostic


class DiagAmplitudeSpectra(Diagnostic):

    def __init__(self):
        super().__init__()
        self.name = 'Amplitude spectra'
        self.tabs = ['fluxtube']

    def get_info(self):
        self.need_file={'field': True,
                        'mom': True}

        return self.need_file
    
    def setup_options(self, run):
        self.specnames = run.parameters.pnt.specnames
        self.time = []
        self.n_fields = run.parameters.pnt.n_fields
        self.n_moms = run.parameters.pnt.n_moms
        self.n_spectra = int(run.parameters.pnt.n_fields + run.parameters.pnt.n_moms*len(self.specnames))
        self.kx=run.spatialgrid.kx_pos
        self.ky=run.spatialgrid.ky
        self.z=run.spatialgrid.z
        self.amplitude_spectra_kx=[]
        self.amplitude_spectra_ky=[]
        self.amplitude_profiles_z=[]
        self.amplitude_profiles_z_ky0=[]
        self.momlist = ["dens", "tpar", "tperp", "qpar", "qperp", "upar", 'densI1', 'TparI1', 'TppI1']
        self.field_mom_names=[]
        self.field_mom_names.append('phi')
        if self.n_fields > 1:
            self.field_mom_names.append('apar')
        if self.n_fields > 2:
            self.field_mom_names.append('bpar')
        for i in self.specnames:
            for j in range(self.n_moms):
                self.field_mom_names.append(self.momlist[j]+i)
        
        
    def execute(self, data, run, steps, extensions,time_point):
        self.time.append(time_point)
        def kx_spectra(a_field):
            #sum ky modes accounting for the complx conj.
            spectra=np.zeros((self.kx.size,a_field.shape[-1]))
            # kx=0 mode
            spectra[0,:] = 2.0*np.sum(np.abs(a_field[0,1:,:])**2,axis=0) + np.abs(a_field[0,0,:])**2
            #finite kx\
            spectra[1:,:] = np.abs(a_field[1:self.kx.size,0,:])**2 + \
                   np.sum(np.abs(a_field[1:self.kx.size,1:,:])**2 + 
                   np.abs(a_field[-1:self.kx.size-2:-1,1:,:])**2,axis=1)
            return spectra
   
        amplitude_kx=np.empty((self.n_spectra, run.spatialgrid.kx_pos.size))
        amplitude_ky=np.empty((self.n_spectra, run.spatialgrid.ky.size))
        amplitude_z=np.empty((self.n_spectra, run.spatialgrid.z.size))
        amplitude_z_ky0=np.empty((self.n_spectra, run.spatialgrid.z.size))
        tmp=data.field.phi(step=steps['field'], extension=extensions['field'])

        amplitude_ky[0,:]=np.average(np.sum(np.abs(tmp)**2,axis=0), weights=run.geometry.jacobian, axis=-1)
        amplitude_kx[0,:]=np.average(kx_spectra(tmp), weights=run.geometry.jacobian, axis=-1)
        amplitude_z[0,:] = xy_av3d(abs(tmp)**2,run.geometry) 
        amplitude_z_ky0[0,:] = np.sum(abs(tmp[:,0,:])**2,axis=0) 
        #plt.plot(self.z,amplitude_z[0,:])
        #plt.plot(self.z,amplitude_z_ky0[0,:])
        #plt.title('amplitude_z')
        #plt.show()
        
        if self.n_fields>1:
            tmp=data.field.A_par(step=steps['field'], extension=extensions['field'])
            amplitude_ky[1,:]=np.average(np.sum(np.abs(tmp)**2,axis=0), weights=run.geometry.jacobian, axis=-1)
            amplitude_kx[1,:]=np.average(kx_spectra(tmp), weights=run.geometry.jacobian, axis=-1)
            amplitude_z[1,:] = xy_av3d(abs(tmp)**2,run.geometry) 
            amplitude_z_ky0[1,:] = np.sum(abs(tmp[:,0,:])**2,axis=0) 
            
        if self.n_fields>2:
            tmp=data.field.B_par(step=steps['field'], extension=extensions['field'])
            amplitude_ky[2,:]=np.average(np.sum(np.abs(tmp)**2,axis=0), weights=run.geometry.jacobian, axis=-1)
            amplitude_kx[2,:]=np.average(kx_spectra(tmp), weights=run.geometry.jacobian, axis=-1)
            amplitude_z[2,:] = xy_av3d(abs(tmp)**2,run.geometry) 
            amplitude_z_ky0[2,:] = np.sum(abs(tmp[:,0,:])**2,axis=0) 
             
        for i_spec, spec in enumerate(run.parameters.specnames):
            for i_quant, quant in enumerate(data.mom[i_spec].varnames.values()):
                tmp = getattr(data.mom[i_spec],quant)(step=steps['mom'], extension=extensions['mom'])
                amplitude_ky[self.n_fields+len(self.specnames)*i_spec+i_quant,:]=np.average(
                        np.sum(np.abs(tmp)**2,axis=0), weights=run.geometry.jacobian, axis=-1)
                amplitude_kx[self.n_fields+len(self.specnames)*i_spec+i_quant,:]=np.average(
                        kx_spectra(tmp), weights=run.geometry.jacobian, axis=-1) 
                amplitude_z[self.n_fields+len(self.specnames)*i_spec+i_quant,:] = xy_av3d(abs(tmp)**2,run.geometry) 
                amplitude_z_ky0[self.n_fields+len(self.specnames)*i_spec+i_quant,:] = np.sum(abs(tmp[:,0,:])**2,axis=0) 
                
        self.amplitude_spectra_ky.append(amplitude_ky)
        self.amplitude_spectra_kx.append(amplitude_kx)
        self.amplitude_profiles_z.append(amplitude_z)
        self.amplitude_profiles_z_ky0.append(amplitude_z_ky0)
                
    def dict_to_mgkdb(self):
        
        amp_spect = {}
        amp_spect['Description'] = 'Spectra of the squares of all available fields and moments as a function of kx or ky (averaged over other variables).  The structure is [field_or_moment][kx or ky][time list][data on the kx or ky grid]. The time points are in [time].  The names of the fields and moments are in [field_mom_names].  The appropriate grids are in [kxgrid] and [kygrid] and  [zgrid].'
        amp_spect['time'] = self.time
        amp_spect['field_mom_names'] = self.field_mom_names
        amp_spect['kxgrid'] = self.kx
        amp_spect['kygrid'] = self.ky
        amp_spect['zgrid'] = self.z
        for i in range(len(self.field_mom_names)):
            amp_spect[self.field_mom_names[i]] = {}
            amp_spect[self.field_mom_names[i]]['kx'] = []
            amp_spect[self.field_mom_names[i]]['ky'] = []
            amp_spect[self.field_mom_names[i]]['z'] = []
            amp_spect[self.field_mom_names[i]]['z_ky0'] = []
            for j in range(len(self.amplitude_spectra_ky)):
                amp_spect[self.field_mom_names[i]]['kx'].append(self.amplitude_spectra_kx[j][i,:])
                amp_spect[self.field_mom_names[i]]['ky'].append(self.amplitude_spectra_ky[j][i,:])
                amp_spect[self.field_mom_names[i]]['z'].append(self.amplitude_profiles_z[j][i,:])
                amp_spect[self.field_mom_names[i]]['z_ky0'].append(self.amplitude_profiles_z_ky0[j][i,:])
        return amp_spect
        
    def plot(self, time_requested, output=None, out_folder=None):
        """ For each selected species we have one figure with six subplots.
            Left is vs. kx, right vs. ky; columnwise we plot, log-log, log-lin, lin-in
            Dashed lines are negative values in log-log plot
            Dashed lines are k multiplied values in log-lin plot"""

        if output:
            output.info_txt.insert(END, "Amplitude specta:\n")

        self.plotbase = Plotting()
        #this is not nice,but kind of only way of not doing it redundantly
        fld_lbl=['phi','apar','bpar']
        mom_lbl=["dens", "tpar", "tperp", "qpar", "qperp", "upar", 'densI1', 'TparI1', 'TppI1']
        
        ''' fields first'''
        fig = plt.figure(figsize=(6, 8))
        ax_loglog_kx = fig.add_subplot(3, 2, 1)
        ax_loglin_kx = fig.add_subplot(3, 2, 3)
        ax_linlin_kx = fig.add_subplot(3, 2, 5)
        ax_loglog_ky = fig.add_subplot(3, 2, 2)
        ax_loglin_ky = fig.add_subplot(3, 2, 4)
        ax_linlin_ky = fig.add_subplot(3, 2, 6)        
        for i_quant in np.arange(self.n_fields):
            if time_requested.size>1:
                ampl_ky = mytrapz(np.asarray(self.amplitude_spectra_ky)[:,i_quant,:], time_requested)
                ampl_kx = mytrapz(np.asarray(self.amplitude_spectra_kx)[:,i_quant,:], time_requested)
            else:
                ampl_ky = np.asarray(self.amplitude_spectra_ky)[0,i_quant,:]
                ampl_kx = np.asarray(self.amplitude_spectra_kx)[0,i_quant,:]
            #print('ky')
            #print(ampl_ky)
            #print('kx')
            #print(ampl_kx)           
            
            # log-log plots, dashed lines for negative values
            baselogkx, = ax_loglog_kx.plot(self.kx, ampl_kx, label=self.plotbase.titles[fld_lbl[i_quant]])
            baselogky, = ax_loglog_ky.plot(self.ky, ampl_ky, label=self.plotbase.titles[fld_lbl[i_quant]])
            # lin-log plots, nothing fancy
            baseloglinkx, = ax_loglin_kx.plot(self.kx, ampl_kx)
            baseloglinky, = ax_loglin_ky.plot(self.ky, ampl_ky)
            # lin-lin plots, nothing fancy
            ax_linlin_kx.plot(self.kx, ampl_kx);
            ax_linlin_ky.plot(self.ky, ampl_ky)

        # set things
        ax_loglog_kx.legend(); ax_loglog_ky.legend()
        ax_loglog_kx.loglog(); ax_loglog_ky.loglog()
        ax_loglin_kx.set_xscale("log"); ax_loglin_ky.set_xscale("log")
        ax_loglin_kx.set_ylim(bottom=0); ax_loglin_ky.set_ylim(bottom=0)
        ax_linlin_kx.set_xlim(left=0); ax_linlin_kx.set_ylim(bottom=0)
        ax_linlin_ky.set_xlim(left=0); ax_linlin_ky.set_ylim(bottom=0)
        ax_linlin_kx.set_xlabel(r"$k_x \rho_{ref}$")
        ax_linlin_ky.set_xlabel(r"$k_y \rho_{ref}$")
        fig.tight_layout()
        fig.show()
  
    
        for i_spec,spec in enumerate(self.specnames):
            fig = plt.figure(figsize=(6, 8))
            ax_loglog_kx = fig.add_subplot(3, 2, 1)
            ax_loglin_kx = fig.add_subplot(3, 2, 3)
            ax_linlin_kx = fig.add_subplot(3, 2, 5)
            ax_loglog_ky = fig.add_subplot(3, 2, 2)
            ax_loglin_ky = fig.add_subplot(3, 2, 4)
            ax_linlin_ky = fig.add_subplot(3, 2, 6)
            
            for i_quant in np.arange(self.n_fields,self.n_fields+self.n_moms):
                if time_requested.size>1:
                    ampl_kx = mytrapz(np.asarray(self.amplitude_spectra_kx)[:,i_quant,:], time_requested)
                    ampl_ky = mytrapz(np.asarray(self.amplitude_spectra_ky)[:,i_quant,:], time_requested)
                else:
                    ampl_kx = np.asarray(self.amplitude_spectra_kx)[0,i_quant,:]
                    ampl_ky = np.asarray(self.amplitude_spectra_ky)[0,i_quant,:]
            
                # log-log plots, dashed lines for negative values
                baselogkx, = ax_loglog_kx.plot(self.kx, ampl_kx, label=self.plotbase.titles[mom_lbl[i_quant-self.n_fields]])
                baselogky, = ax_loglog_ky.plot(self.ky, ampl_ky, label=self.plotbase.titles[mom_lbl[i_quant-self.n_fields]])
                # lin-log plots, nothing fancy
                baseloglinkx, = ax_loglin_kx.plot(self.kx, ampl_kx)
                baseloglinky, = ax_loglin_ky.plot(self.ky, ampl_ky)
                # lin-lin plots, nothing fancy
                ax_linlin_kx.plot(self.kx, ampl_kx)
                ax_linlin_ky.plot(self.ky, ampl_ky)

            # set things
            ax_loglog_kx.loglog(); ax_loglog_ky.loglog()
            ax_loglog_kx.legend(); ax_loglog_ky.legend()
            ax_loglin_kx.set_xscale("log"); ax_loglin_ky.set_xscale("log")
            ax_loglin_kx.set_ylim(bottom=0); ax_loglin_ky.set_ylim(bottom=0)
            ax_linlin_kx.set_xlim(left=0); ax_linlin_kx.set_ylim(bottom=0)
            ax_linlin_ky.set_xlim(left=0); ax_linlin_ky.set_ylim(bottom=0)
            ax_linlin_kx.set_xlabel(r"$k_x \rho_{ref}$")
            ax_linlin_ky.set_xlabel(r"$k_y \rho_{ref}$")
            ax_loglog_kx.set_title("{}".format(spec))
            fig.tight_layout()
            fig.show()



""" Classes for plotting field based mode structures """

import matplotlib.pyplot as plt
import numpy as np
import warnings
from multiprocessing.dummy import Pool

from .baseplot import Plotting
from ..data.slices import MomFieldSlice
from ..utils.comm import DiagSpace

from ..utils.averages import mytrapz
from ..utils.geom import Geometry


class ModeStructure(object):
    """ Ballooning Mode Structure for selected ky of field data """

    def __init__(self, common, rundatafiles, species=None, moms=None, kyind=None, \
                 normalize=False, tavg=True):
        if not common.x_local or not common.y_local:
            raise NotImplementedError("Not yet implemented for nonlocal simulations")
        self.cm = common
        self.species = species
        self.timearray = None
        self.normalize = normalize

        if moms is None:
            self.ballamps = {"phi": []}
            if common.electromagnetic:
                self.ballamps.update({"apar": []})
                if common.bpar:
                    self.ballamps.update({"bpar": []})
        else:
            self.ballamps = {}
            for mom in moms:
                self.ballamps.update({mom : []})

        if kyind is None:
            self.kyind=(1 if self.cm.pnt.nonlinear else 0) #default: 1st finite ky
        else:
            self.kyind=kyind[0]

        self.calc_modestructure(species=species,rundatafiles=rundatafiles,tavg=tavg)

    def _fetch_moms(self, rundatafiles, species):
        """ Get mom and field slice object for species"""
        diagspace = DiagSpace(self.cm.spatialgrid, x_fourier=True, y_fourier=True)

        momamps = {}
        for mom in self.ballamps.keys():
            momamps.update({mom : None})

        with Pool() as p:
            momamp_temp = p.map(
                lambda mom: MomFieldSlice(self.cm, mom, species, diagspace, rundatafiles,
                                          modifier_func=lambda dum: dum), momamps)
        for imom, mom in enumerate(momamps):
            momamps[mom] = momamp_temp[imom]
        return momamps

    def calc_modestructure(self,species,rundatafiles,tavg):
        """ Calculate the mode structures"""
        momamps = self._fetch_moms(rundatafiles, species=species)
        geom = Geometry(self.cm)
        timerefvar = list(self.ballamps.keys())[0]  #TODO: NEED TO SET TO LOWEST TIME RES. VAR/FILE
        momamps[timerefvar].check_times()
        pos = momamps[timerefvar].calc_positions()
        self.timearray = np.take(momamps[timerefvar].timearray, pos)

        if self.cm.pnt.magn_geometry == "'s_alpha'" or self.cm.pnt.magn_geometry == "'circular'":
            Cyq0_x0 = 1.0
        else:
            Cyq0_x0 = geom.Cy * self.cm.pnt.q0 / self.cm.pnt.x0
        sign_shear = (-1 if self.cm.pnt.shat < 0 else 1)
        nexc_sign = sign_shear*self.cm.pnt.sign_Ip_CW*self.cm.pnt.sign_Bt_CW

        if self.cm.pnt.adapt_lx:
            nexc = 1
        else:
            nexc = int(np.round(self.cm.pnt.lx*self.cm.pnt.n_pol*
                    np.absolute(self.cm.pnt.shat)*self.cm.pnt.kymin*
                    np.absolute(Cyq0_x0)))
        nexc*=nexc_sign
        nexcj = nexc*(self.cm.pnt.ky0_ind+self.kyind)
        nconn = int(int(int(self.cm.pnt.nx0-1)/2)/np.abs(nexcj))*2+1
        phasefac = (-1)**nexcj*np.exp(2.0*np.pi*1j*self.cm.pnt.n0_global*
                                         self.cm.pnt.q0*(self.cm.pnt.ky0_ind+self.kyind))

        self.chi_ext = []
        for ncind in range(-int(nconn/2),int(nconn/2)+1):
            self.chi_ext = np.concatenate([self.chi_ext,
                                           (self.cm.spatialgrid.z+ncind*2.0*np.pi)])
        for time in self.timearray:
            def mom_t(momname):
                return momamps[momname].generate_slice_attime(time)
            for mom in momamps:
                if self.normalize:
                    normval = mom_t(mom)[0,self.kyind,int(self.cm.pnt.nz0/2)]
                else:
                    normval = 0.
                balloon = []
                for ncind in range(-int(nconn/2),int(nconn/2)+1):
                    balloon = np.concatenate([balloon,np.conj(phasefac)**ncind*
                                              mom_t(mom)[ncind*nexcj,self.kyind]])
                if normval == 0.:
                    normphase=0.
                    normval=1.
                else:
                    normphase = np.angle(normval)
                    normval=np.absolute(normval)
                self.ballamps[mom].append(balloon*np.exp(-1j*normphase)/normval)

        if tavg:
            for mom in self.ballamps:
                self.ballamps[mom] = mytrapz(self.ballamps[mom], self.timearray)


class PlotModeStructure(Plotting):

    def __init__(self, commonlist, ballserieslist):
        super().__init__()
        self.balllist = ballserieslist
        self.cm = commonlist
        self.titles.update({"phi": r"$\phi$", "apar": r"$A_\parallel$", "bpar": r"$B_\parallel$"})

    def createfigs(self):
        for ballseries in self.balllist:
            moms = ballseries.ballamps.keys()
            fig = plt.figure("Ballooning Mode Structure")

            chi_pi = ballseries.chi_ext/np.pi
            kyval = ballseries.cm.spatialgrid.ky[ballseries.kyind]

            for imom,mom in enumerate(moms):
                ax_lin = fig.add_subplot(1,len(moms),imom+1)
                ballarr = ballseries.ballamps[mom]
                ax_lin.plot(chi_pi, np.absolute(ballarr),
                            label="|{}|".format(self.titles[mom]),
                            color='black')
                ax_lin.plot(chi_pi, np.real(ballarr),
                            label="Re({})".format(self.titles[mom]),
                            color='blue',lw=1)
                ax_lin.plot(chi_pi, np.imag(ballarr),
                            label="Im({})".format(self.titles[mom]),
                            color='red',lw=1)
                ax_lin.set_xlabel(r"$\chi/\pi$")
                ax_lin.set_title(r"{0:s}$(k_y\rho={1:6.3f})$".
                                 format(self.titles[mom],kyval))
                ax_lin.axhline(y=0,color='black',lw=1,ls='--')
                ax_lin.legend()

            fig.tight_layout()



# -*- coding: utf-8 -*-
""" Module for reading omega and eigenvalue output files """
import numpy as np

class Eigenvaluedata(object):
    """ Collect the data of an omega or eigenvalue file and store it in a nice numpy array"""

    def __init__(self,common):
        if common.pnt.nonlinear:
            raise RuntimeError("Not a linear simulation - won't read eigenvalues")

        if common.pars["comp_type"]=="'IV'":
            self.parseomegafile(common.fileextension)
        else:
            self.parseeigenvaluesfile(common.fileextension)

    def parseomegafile(self,fileextension):
        """ Parse omega.dat """
        try:
            with open("omega"+fileextension, 'r') as omegafile:
                data = omegafile.readline().split()
                if len(data) < 3:
                    data = data + [np.nan for _ in range(3-len(data))]
                self.growth_rate = [float(data[-2])]
                self.frequency = [float(data[-1])]
        except IOError:
            print("Could not read omega"+fileextension+" file."+ "Fill with NaN!")
            self.growth_rate = [np.nan]
            self.frequency = [np.nan]
#            raise

    def parseeigenvaluesfile(self,fileextension):
        """ Parse eigenvalues.dat """
        try:
            self.growth_rate = []
            self.frequency = []
            with open("eigenvalues"+fileextension, 'r') as evfile:
                header = evfile.readline()
                for line in evfile:
                    data = line.split()
                    self.growth_rate += [float(data[-2])]
                    self.frequency += [float(data[-1])]
        except IOError:
            print("Could not read eigenvalues"+fileextension+" file")
            raise
